#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
import argparse
import ast
import contextlib
import pprint
import random
import os
import termios
import timeit
import tty

import gym

import nle  # noqa: F401
from nle import nethack
from nle.nethack import print_message


@contextlib.contextmanager
def dummy_context():
    yield None


@contextlib.contextmanager
def no_echo():
    tt = termios.tcgetattr(0)
    try:
        tty.setraw(0)
        yield
    finally:
        termios.tcsetattr(0, termios.TCSAFLUSH, tt)


def get_action(env, action_mode, is_raw_env):
    if action_mode == "random":
        if not is_raw_env:
            action = env.action_space.sample()
        else:
            action = random.choice(nle.env.base.FULL_ACTIONS)
    elif action_mode == "human":
        while True:
            with no_echo():
                ch = ord(os.read(0, 1))
            if ch in [nethack.C("c"), ord(b"q")]:
                print("Received exit code {}. Aborting.".format(ch))
                return None
            try:
                action = env._actions.index(ch)
                break
            except ValueError:
                print(
                    ("Selected action '%s' is not in action list. Please try again.")
                    % chr(ch)
                )
                continue
    return action


def play(env_name, play_mode, ngames, max_steps, seeds, no_clear, no_render):
    is_raw_env = env_name == "nethack"

    if is_raw_env:
        # TODO save data somewhere reasonable
        env = nethack.NetHack(archivefile="./nle_data/play_data.zip")
    else:
        env = gym.make(env_name)
        if seeds is not None:
            env.seed(seeds)

    obs = env.reset()
    last_obs = None  # needed for "nethack" env

    steps = 0
    episodes = 0
    reward = 0.0
    action = None
    ch = None

    start_time = timeit.default_timer()
    while True:
        if not no_render:
            if not no_clear:
                os.system("cls" if os.name == "nt" else "clear")

            if not is_raw_env:
                print("Previous reward:", reward)
                print("Available actions:", env._actions)
                print(
                    "Previous action: {}{!r})".format(
                        "{} --".format(chr(ch)) if ch is not None else "",
                        env._actions[action] if action is not None else None,
                    )
                )
                env.render()
            else:
                print("Available actions:", nle.env.base.FULL_ACTIONS)
                print("Previous actions:", action)
                print_message.print_message(obs)

        action = get_action(env, play_mode, is_raw_env)
        if action is None:
            break

        if is_raw_env:
            last_obs = obs
            obs, done, info = env.step(action)
        else:
            obs, reward, done, info = env.step(action)
        steps += 1

        done = done or steps >= max_steps
        if not done:
            continue

        if not is_raw_env:
            print("Final reward:", reward)
            print("End status:", info["end_status"].name)
            print("Env stats:")
            pprint.pprint(info["stats"])
        else:
            if last_obs.ProgramState().Gameover():
                # Print tombstone.
                if last_obs.WindowsLength() < 1:
                    return steps
                window = last_obs.Windows(1)

                if not no_render:
                    for i in range(window.StringsLength()):
                        print(window.Strings(i).decode("ascii"))

        time_delta = timeit.default_timer() - start_time
        print(
            "Episode: {}. Steps: {}. SPS: {:f}".format(
                episodes, steps, steps / time_delta
            )
        )
        start_time = timeit.default_timer()

        episodes += 1
        steps = 0
        if episodes == ngames:
            break
        env.reset()


def main():
    parser = argparse.ArgumentParser(description="NLE Play tool.")
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Enables debug mode, which will drop stack into "
        "an ipdb shell if an exception is raised.",
    )
    parser.add_argument(
        "-m",
        "--mode",
        type=str,
        default="human",
        choices=["human", "random"],
        help="Control mode.",
    )
    parser.add_argument(
        "-e",
        "--env",
        type=str,
        default="NetHackStaircase-v0",
        help="Gym environment spec.",
    )
    parser.add_argument(
        "-n",
        "--ngames",
        type=int,
        default=1,
        help="Number of games to be played before exiting. "
        "NetHack will auto-restart if > 1.",
    )
    parser.add_argument(
        "--max-steps",
        type=int,
        default=10000,
        help="Number of maximum steps per episode.",
    )
    parser.add_argument(
        "--seeds",
        default=None,
        help="Seeds to send to NetHack. Can be a dict or int. "
        "Defaults to None (no seeding).",
    )

    parser.add_argument("--no-clear", action="store_true", help="Disables tty clears.")
    parser.add_argument(
        "--no-render", action="store_true", help="Disables env.render()."
    )
    flags = parser.parse_args()

    if flags.debug:
        import ipdb

        cm = ipdb.launch_ipdb_on_exception
    else:
        cm = dummy_context

    with cm():
        if flags.seeds is not None:
            # to handle both int and dicts
            flags.seeds = ast.literal_eval(flags.seeds)

        play(
            flags.env,
            flags.mode,
            flags.ngames,
            flags.max_steps,
            flags.seeds,
            flags.no_clear,
            flags.no_render,
        )


if __name__ == "__main__":
    main()



# Copyright (c) Facebook, Inc. and its affiliates.
import gc
import weakref

import gym

import nle  # noqa: F401


def get_ref_objects(exclude=None):
    result = {}
    for o in gc.get_objects():
        if o is exclude:
            continue
        try:
            result[id(o)] = weakref.ref(o)
        except TypeError:
            pass
    return result


def play(env):
    env.reset()
    for _ in range(50):
        _, _, done, _ = env.step(env.action_space.sample())
        if done:
            break


class TestObjectLifetime:
    def test_restart(self):
        env = gym.make("NetHack-v0")
        play(env)
        play(env)

    def test_object_allocation(self):
        env = gym.make("NetHack-v0")
        play(env)

        before = get_ref_objects()

        for _ in range(10):
            play(env)

        after = get_ref_objects(before)
        assert len(before) == len(after)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Condition(object):
    __slots__ = ['_tab']

    # Condition
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Condition
    def STONE(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Condition
    def SLIME(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(1))
    # Condition
    def STRNGL(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(2))
    # Condition
    def FOODPOIS(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(3))
    # Condition
    def TERMILL(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4))
    # Condition
    def BLIND(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(5))
    # Condition
    def DEAF(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(6))
    # Condition
    def STUN(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(7))
    # Condition
    def CONF(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))
    # Condition
    def HALLU(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(9))
    # Condition
    def LEV(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(10))
    # Condition
    def FLY(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(11))
    # Condition
    def RIDE(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(12))

def CreateCondition(builder, STONE, SLIME, STRNGL, FOODPOIS, TERMILL, BLIND, DEAF, STUN, CONF, HALLU, LEV, FLY, RIDE):
    builder.Prep(1, 13)
    builder.PrependBool(RIDE)
    builder.PrependBool(FLY)
    builder.PrependBool(LEV)
    builder.PrependBool(HALLU)
    builder.PrependBool(CONF)
    builder.PrependBool(STUN)
    builder.PrependBool(DEAF)
    builder.PrependBool(BLIND)
    builder.PrependBool(TERMILL)
    builder.PrependBool(FOODPOIS)
    builder.PrependBool(STRNGL)
    builder.PrependBool(SLIME)
    builder.PrependBool(STONE)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class MenuItem(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsMenuItem(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = MenuItem()
        x.Init(buf, n + offset)
        return x

    # MenuItem
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # MenuItem
    def Glyph(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

    # MenuItem
    def Selector(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MenuItem
    def Gselector(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MenuItem
    def Str(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # MenuItem
    def Selected(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def MenuItemStart(builder): builder.StartObject(5)
def MenuItemAddGlyph(builder, glyph): builder.PrependInt16Slot(0, glyph, 0)
def MenuItemAddSelector(builder, selector): builder.PrependInt8Slot(1, selector, 0)
def MenuItemAddGselector(builder, gselector): builder.PrependInt8Slot(2, gselector, 0)
def MenuItemAddStr(builder, str): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(str), 0)
def MenuItemAddSelected(builder, selected): builder.PrependBoolSlot(4, selected, 0)
def MenuItemEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Window(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsWindow(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Window()
        x.Init(buf, n + offset)
        return x

    # Window
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Window
    def Type(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Window
    def MenuItems(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from nle.fbs.MenuItem import MenuItem
            obj = MenuItem()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Window
    def MenuItemsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Window
    def MenuItemsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Window
    def Strings(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Window
    def StringsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Window
    def StringsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def WindowStart(builder): builder.StartObject(3)
def WindowAddType(builder, type): builder.PrependInt8Slot(0, type, 0)
def WindowAddMenuItems(builder, menuItems): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(menuItems), 0)
def WindowStartMenuItemsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def WindowAddStrings(builder, strings): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(strings), 0)
def WindowStartStringsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def WindowEnd(builder): return builder.EndObject()



# Copyright (c) Facebook, Inc. and its affiliates.
import numpy as np


PROGRAM_STATE_FIELDS = [
    "Gameover",
    "Panicking",
    "Exiting",
    "InMoveloop",
    "InImpossible",
]

STATUS_FIELDS = [
    "TITLE",
    "STR",
    "DX",
    "CO",
    "IN",
    "WI",
    "CH",  # 1..6
    "ALIGN",
    "SCORE",
    "CAP",
    "GOLD",
    "ENE",
    "ENEMAX",  # 7..12
    "XP",
    "AC",
    "HD",
    "TIME",
    "HUNGER",
    "HP",
    "HPMAX",
    "LEVELDESC",  # 13..20
    "EXP",
]

CONDITION_FIELDS = [
    "STONE",
    "SLIME",
    "STRNGL",
    "FOODPOIS",
    "TERMILL",
    "BLIND",
    "DEAF",
    "STUN",
    "CONF",
    "HALLU",
    "LEV",
    "FLY",
    "RIDE",
]

BLSTATS_FIELDS = [
    "CursX",
    "CursY",
    "StrengthPercentage",
    "Strength",
    "Dexterity",
    "Constitution",
    "Intelligence",
    "Wisdom",
    "Charisma",
    "Score",
    "Hitpoints",
    "MaxHitpoints",
    "Depth",
    "Gold",
    "Energy",
    "MaxEnergy",
    "ArmorClass",
    "MonsterLevel",
    "ExperienceLevel",
    "ExperiencePoints",
    "Time",
    "HungerState",
    "CarryingCapacity",
]


INVENTORYITEM_FIELDS = ["Glyph", "Str", "Letter", "ObjectClass", "ObjectClassName"]

MENUITEM_FIELDS = ["Glyph", "Selector", "Gselector", "Str", "Selected"]

WIN_MESSAGE = 1  # Technically dynamic. Practically constant.

SEEDS_FIELDS = ["Core", "Disp"]


def fb_ndarray_to_np(fb_ndarray):
    result = fb_ndarray.DataAsNumpy()
    result = result.view(np.typeDict[fb_ndarray.Dtype()])
    result = result.reshape(fb_ndarray.ShapeAsNumpy().tolist())
    return result


def print_message(message):
    fb_windows = [message.Windows(i) for i in range(message.WindowsLength())]

    for fb_window in fb_windows:
        window = {}
        window["strings"] = [
            fb_window.Strings(i) for i in range(fb_window.StringsLength())
        ]
        window["items"] = []
        fb_items = [fb_window.MenuItems(i) for i in range(fb_window.MenuItemsLength())]
        for fb_item in fb_items:
            window["items"].append(
                {field: getattr(fb_item, field)() for field in MENUITEM_FIELDS}
            )
        window["type"] = fb_window.Type()
        print("window", window)

    seeds = {k: getattr(message.Seeds(), k)() for k in SEEDS_FIELDS}
    print("seeds", seeds)

    if message.NotRunning():
        return "done"

    program_state = message.ProgramState()
    program_state_dict = {
        field: getattr(program_state, field)() for field in PROGRAM_STATE_FIELDS
    }
    print("program_state", program_state_dict)

    if not program_state.InMoveloop():
        print("Game not in move loop.")
        return

    obs = message.Observation()
    status = obs.Status()
    status_dict = {
        field: getattr(status, field)().decode("utf-8") for field in STATUS_FIELDS
    }
    condition = status.Condition()
    condition_dict = {field: getattr(condition, field)() for field in CONDITION_FIELDS}
    print("status", status_dict)
    print("condition", condition_dict)

    blstats = message.Blstats()
    blstats_dict = {field: getattr(blstats, field)() for field in BLSTATS_FIELDS}
    print("blstats", blstats_dict)

    inventory = [obs.Inventory(i) for i in range(obs.InventoryLength())]
    items = [
        {field: getattr(item, field)() for field in INVENTORYITEM_FIELDS}
        for item in inventory
    ]
    print("inv_items", items)
    for item in items:
        print(chr(item["Letter"]), item["Str"])

    internal = message.Internal()
    print("deepest_lev_reached", internal.DeepestLevReached())
    if internal.KillerName():
        print("killer_name", internal.KillerName())
    call_stack = [internal.CallStack(i) for i in range(internal.CallStackLength())]
    if internal.Xwaitforspace():
        print("Waiting for space/return")
    print("call_stack", call_stack)
    if internal.StairsDown():
        print("On downward stairs")

    message_win = message.Windows(WIN_MESSAGE)
    messages = [message_win.Strings(i) for i in range(message_win.StringsLength())]
    if messages:
        print("Messages:")
        for m in messages:
            print(m)

    chars = fb_ndarray_to_np(obs.Chars())
    colors = fb_ndarray_to_np(obs.Colors())
    rows, cols = chars.shape
    nh_HE = "\033[0m"
    BRIGHT = 8
    for r in range(rows):
        if False:  # no colors.
            print(chars[r].tobytes().decode("utf-8"))
            continue
        for c in range(cols):
            # cf. termcap.c.
            start_color = "\033[%d" % bool(colors[r][c] & BRIGHT)
            start_color += ";3%d" % (colors[r][c] & ~BRIGHT)
            start_color += "m"

            print(start_color + chr(chars[r][c]), end=nh_HE)
        print("")



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
# This script outputs relevant system environment info
# Run it with `python collect_env.py`.
# Original version from:
# https://github.com/pytorch/pytorch/blob/fa5bc9fa/torch/utils/collect_env.py

from __future__ import absolute_import, division, print_function, unicode_literals
import locale
import re
import subprocess
import sys
import os
from collections import namedtuple

try:
    import nle

    NLE_AVAILABLE = True
except (ImportError, NameError, AttributeError):
    NLE_AVAILABLE = False

try:
    import torch

    TORCH_AVAILABLE = True
except (ImportError, NameError, AttributeError):
    TORCH_AVAILABLE = False

PY3 = sys.version_info >= (3, 0)

# System Environment Information
SystemEnv = namedtuple(
    "SystemEnv",
    [
        "nle_version",
        "torch_version",
        "is_debug_build",
        "cuda_compiled_version",
        "gcc_version",
        "cmake_version",
        "os",
        "python_version",
        "is_cuda_available",
        "cuda_runtime_version",
        "nvidia_driver_version",
        "nvidia_gpu_models",
        "cudnn_version",
        "pip_version",  # 'pip' or 'pip3'
        "pip_packages",
        "conda_packages",
    ],
)


def run(command):
    """Returns (return-code, stdout, stderr)"""
    p = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
    )
    output, err = p.communicate()
    rc = p.returncode
    if PY3:
        enc = locale.getpreferredencoding()
        output = output.decode(enc)
        err = err.decode(enc)
    return rc, output.strip(), err.strip()


def run_and_read_all(run_lambda, command):
    """Runs command using run_lambda; reads and returns entire output if rc is 0"""
    rc, out, _ = run_lambda(command)
    if rc != 0:
        return None
    return out


def run_and_parse_first_match(run_lambda, command, regex):
    """Runs command using run_lambda, returns the first regex match if it exists"""
    rc, out, _ = run_lambda(command)
    if rc != 0:
        return None
    match = re.search(regex, out)
    if match is None:
        return None
    return match.group(1)


def get_conda_packages(run_lambda):
    if get_platform() == "win32":
        grep_cmd = r'findstr /R "torch soumith mkl magma"'
    else:
        grep_cmd = r'grep "torch\|soumith\|mkl\|magma"'
    conda = os.environ.get("CONDA_EXE", "conda")
    out = run_and_read_all(run_lambda, conda + " list | " + grep_cmd)
    if out is None:
        return out
    # Comment starting at beginning of line
    comment_regex = re.compile(r"^#.*\n")
    return re.sub(comment_regex, "", out)


def get_gcc_version(run_lambda):
    return run_and_parse_first_match(run_lambda, "gcc --version", r"gcc (.*)")


def get_cmake_version(run_lambda):
    return run_and_parse_first_match(run_lambda, "cmake --version", r"cmake (.*)")


def get_nvidia_driver_version(run_lambda):
    if get_platform() == "darwin":
        cmd = "kextstat | grep -i cuda"
        return run_and_parse_first_match(
            run_lambda, cmd, r"com[.]nvidia[.]CUDA [(](.*?)[)]"
        )
    smi = get_nvidia_smi()
    return run_and_parse_first_match(run_lambda, smi, r"Driver Version: (.*?) ")


def get_gpu_info(run_lambda):
    if get_platform() == "darwin":
        if TORCH_AVAILABLE and torch.cuda.is_available():
            return torch.cuda.get_device_name(None)
        return None
    smi = get_nvidia_smi()
    uuid_regex = re.compile(r" \(UUID: .+?\)")
    rc, out, _ = run_lambda(smi + " -L")
    if rc != 0:
        return None
    # Anonymize GPUs by removing their UUID
    return re.sub(uuid_regex, "", out)


def get_running_cuda_version(run_lambda):
    return run_and_parse_first_match(run_lambda, "nvcc --version", r"V(.*)$")


def get_cudnn_version(run_lambda):
    """This will return a list of libcudnn.so

    It's hard to tell which one is being used.
    """
    if get_platform() == "win32":
        cudnn_cmd = 'where /R "%CUDA_PATH%\\bin" cudnn*.dll'
    elif get_platform() == "darwin":
        # CUDA libraries and drivers can be found in /usr/local/cuda/. See
        # https://docs.nvidia.com/cuda/cuda-installation-guide-mac-os-x/index.html#install # noqa
        # https://docs.nvidia.com/deeplearning/sdk/cudnn-install/index.html#installmac # noqa
        # Use CUDNN_LIBRARY when cudnn library is installed elsewhere.
        cudnn_cmd = "ls /usr/local/cuda/lib/libcudnn*"
    else:
        cudnn_cmd = 'ldconfig -p | grep libcudnn | rev | cut -d" " -f1 | rev'
    rc, out, _ = run_lambda(cudnn_cmd)
    # find will return 1 if there are permission errors or if not found
    if len(out) == 0 or (rc != 1 and rc != 0):
        cl = os.environ.get("CUDNN_LIBRARY")
        if cl is not None and os.path.isfile(cl):
            return os.path.realpath(cl)
        return None
    files = set()
    for fn in out.split("\n"):
        fn = os.path.realpath(fn)  # eliminate symbolic links
        if os.path.isfile(fn):
            files.add(fn)
    if not files:
        return None
    # Alphabetize the result because the order is non-deterministic otherwise
    files = list(sorted(files))
    if len(files) == 1:
        return files[0]
    result = "\n".join(files)
    return "Probably one of the following:\n{}".format(result)


def get_nvidia_smi():
    # Note: nvidia-smi is currently available only on Windows and Linux
    smi = "nvidia-smi"
    if get_platform() == "win32":
        smi = '"C:\\Program Files\\NVIDIA Corporation\\NVSMI\\%s"' % smi
    return smi


def get_platform():
    if sys.platform.startswith("linux"):
        return "linux"
    elif sys.platform.startswith("win32"):
        return "win32"
    elif sys.platform.startswith("cygwin"):
        return "cygwin"
    elif sys.platform.startswith("darwin"):
        return "darwin"
    else:
        return sys.platform


def get_mac_version(run_lambda):
    return run_and_parse_first_match(run_lambda, "sw_vers -productVersion", r"(.*)")


def get_windows_version(run_lambda):
    return run_and_read_all(run_lambda, "wmic os get Caption | findstr /v Caption")


def get_lsb_version(run_lambda):
    return run_and_parse_first_match(
        run_lambda, "lsb_release -a", r"Description:\t(.*)"
    )


def check_release_file(run_lambda):
    return run_and_parse_first_match(
        run_lambda, "cat /etc/*-release", r'PRETTY_NAME="(.*)"'
    )


def get_os(run_lambda):
    platform = get_platform()

    if platform == "win32" or platform == "cygwin":
        return get_windows_version(run_lambda)

    if platform == "darwin":
        version = get_mac_version(run_lambda)
        if version is None:
            return None
        return "Mac OSX {}".format(version)

    if platform == "linux":
        # Ubuntu/Debian based
        desc = get_lsb_version(run_lambda)
        if desc is not None:
            return desc

        # Try reading /etc/*-release
        desc = check_release_file(run_lambda)
        if desc is not None:
            return desc

        return platform

    # Unknown platform
    return platform


def get_pip_packages(run_lambda):
    # People generally have `pip` as `pip` or `pip3`
    def run_with_pip(pip):
        if get_platform() == "win32":
            grep_cmd = r'findstr /R "numpy torch"'
        else:
            grep_cmd = r'grep "torch\|numpy"'
        return run_and_read_all(run_lambda, pip + " list --format=freeze | " + grep_cmd)

    if not PY3:
        return "pip", run_with_pip("pip")

    # Try to figure out if the user is running pip or pip3.
    out2 = run_with_pip("pip")
    out3 = run_with_pip("pip3")

    num_pips = len([x for x in [out2, out3] if x is not None])
    if num_pips == 0:
        return "pip", out2

    if num_pips == 1:
        if out2 is not None:
            return "pip", out2
        return "pip3", out3

    # num_pips is 2. Return pip3 by default b/c that most likely
    # is the one associated with Python 3
    return "pip3", out3


def get_env_info():
    run_lambda = run
    pip_version, pip_list_output = get_pip_packages(run_lambda)

    if NLE_AVAILABLE:
        nle_version = nle.__version__
    else:
        nle_version = "N/A"

    if TORCH_AVAILABLE:
        torch_version_str = torch.__version__
        torch_debug_mode_str = torch.version.debug
        cuda_available_str = torch.cuda.is_available()
        cuda_version_str = torch.version.cuda
    else:
        torch_version_str = (
            torch_debug_mode_str
        ) = cuda_available_str = cuda_version_str = "N/A"

    return SystemEnv(
        nle_version=nle_version,
        torch_version=torch_version_str,
        is_debug_build=torch_debug_mode_str,
        python_version="{}.{}".format(sys.version_info[0], sys.version_info[1]),
        is_cuda_available=cuda_available_str,
        cuda_compiled_version=cuda_version_str,
        cuda_runtime_version=get_running_cuda_version(run_lambda),
        nvidia_gpu_models=get_gpu_info(run_lambda),
        nvidia_driver_version=get_nvidia_driver_version(run_lambda),
        cudnn_version=get_cudnn_version(run_lambda),
        pip_version=pip_version,
        pip_packages=pip_list_output,
        conda_packages=get_conda_packages(run_lambda),
        os=get_os(run_lambda),
        gcc_version=get_gcc_version(run_lambda),
        cmake_version=get_cmake_version(run_lambda),
    )


env_info_fmt = """
NLE version: {nle_version}
PyTorch version: {torch_version}
Is debug build: {is_debug_build}
CUDA used to build PyTorch: {cuda_compiled_version}

OS: {os}
GCC version: {gcc_version}
CMake version: {cmake_version}

Python version: {python_version}
Is CUDA available: {is_cuda_available}
CUDA runtime version: {cuda_runtime_version}
GPU models and configuration: {nvidia_gpu_models}
Nvidia driver version: {nvidia_driver_version}
cuDNN version: {cudnn_version}

Versions of relevant libraries:
{pip_packages}
{conda_packages}
""".strip()


def pretty_str(envinfo):
    def replace_nones(dct, replacement="Could not collect"):
        for key in dct.keys():
            if dct[key] is not None:
                continue
            dct[key] = replacement
        return dct

    def replace_bools(dct, true="Yes", false="No"):
        for key in dct.keys():
            if dct[key] is True:
                dct[key] = true
            elif dct[key] is False:
                dct[key] = false
        return dct

    def prepend(text, tag="[prepend]"):
        lines = text.split("\n")
        updated_lines = [tag + line for line in lines]
        return "\n".join(updated_lines)

    def replace_if_empty(text, replacement="No relevant packages"):
        if text is not None and len(text) == 0:
            return replacement
        return text

    def maybe_start_on_next_line(string):
        # If `string` is multiline, prepend a \n to it.
        if string is not None and len(string.split("\n")) > 1:
            return "\n{}\n".format(string)
        return string

    mutable_dict = envinfo._asdict()

    # If nvidia_gpu_models is multiline, start on the next line
    mutable_dict["nvidia_gpu_models"] = maybe_start_on_next_line(
        envinfo.nvidia_gpu_models
    )

    # If the machine doesn't have CUDA, report some fields as 'No CUDA'
    dynamic_cuda_fields = [
        "cuda_runtime_version",
        "nvidia_gpu_models",
        "nvidia_driver_version",
    ]
    all_cuda_fields = dynamic_cuda_fields + ["cudnn_version"]
    all_dynamic_cuda_fields_missing = all(
        mutable_dict[field] is None for field in dynamic_cuda_fields
    )
    if (
        TORCH_AVAILABLE
        and not torch.cuda.is_available()
        and all_dynamic_cuda_fields_missing
    ):
        for field in all_cuda_fields:
            mutable_dict[field] = "No CUDA"
        if envinfo.cuda_compiled_version is None:
            mutable_dict["cuda_compiled_version"] = "None"

    # Replace True with Yes, False with No
    mutable_dict = replace_bools(mutable_dict)

    # Replace all None objects with 'Could not collect'
    mutable_dict = replace_nones(mutable_dict)

    # If either of these are '', replace with 'No relevant packages'
    mutable_dict["pip_packages"] = replace_if_empty(mutable_dict["pip_packages"])
    mutable_dict["conda_packages"] = replace_if_empty(mutable_dict["conda_packages"])

    # Tag conda and pip packages with a prefix
    # If they were previously None, they'll show up as ie '[conda] Could not collect'
    if mutable_dict["pip_packages"]:
        mutable_dict["pip_packages"] = prepend(
            mutable_dict["pip_packages"], "[{}] ".format(envinfo.pip_version)
        )
    if mutable_dict["conda_packages"]:
        mutable_dict["conda_packages"] = prepend(
            mutable_dict["conda_packages"], "[conda] "
        )
    return env_info_fmt.format(**mutable_dict)


def get_pretty_env_info():
    return pretty_str(get_env_info())


def main():
    print("Collecting environment information...")
    output = get_pretty_env_info()
    print(output)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
import random
import sys

import numpy as np
import pytest
import gym

import nle
import nle.env


def get_nethack_env_ids():
    specs = gym.envs.registry.all()
    # Ignoring base environment, since we can't handle random actions yet with
    # the full action space, and this requires a whole different set of tests.
    # For now this is OK, since NetHackScore-v0 is very similar.
    return [
        spec.id
        for spec in specs
        if spec.id.startswith("NetHack") and spec.id != "NetHack-v0"
    ]


def rollout_env(env, max_rollout_len):
    """Produces a rollout and asserts step outputs.

    Does *not* assume that the environment has already been reset.
    """
    obs = env.reset()
    assert env.observation_space.contains(obs)

    step = 0
    while True:
        a = env.action_space.sample()
        obs, reward, done, info = env.step(a)
        assert env.observation_space.contains(obs)
        assert isinstance(reward, float)
        assert isinstance(done, bool)
        assert isinstance(info, dict)
        if done or step >= max_rollout_len:
            break
    env.close()


def compare_rollouts(env0, env1, max_rollout_len):
    """Checks that two active environments return the same rollout.

    Assumes that the environments have already been reset.
    """
    step = 0
    while True:
        a = env0.action_space.sample()
        obs0, reward0, done0, info0 = env0.step(a)
        obs1, reward1, done1, info1 = env1.step(a)
        step += 1
        np.testing.assert_equal(obs0, obs1)
        assert reward0 == reward1
        assert done0 == done1

        # pid entries (and thus ttyrecs) won't match. Copy before removing.
        info0, info1 = info0.copy(), info1.copy()
        del info0["pid"], info1["pid"]

        if done0:
            assert "stats" in info0  # just to be sure
            assert "stats" in info1

            for k in ["ttyrec"]:
                assert info0["stats"][k] != info1["stats"][k]

                del info0["stats"][k]
                del info1["stats"][k]

        assert info0 == info1

        if done0 or step >= max_rollout_len:
            return


@pytest.mark.parametrize("env_name", get_nethack_env_ids())
class TestGymEnv:
    @pytest.yield_fixture(autouse=True)  # will be applied to all tests in class
    def make_cwd_tmp(self, tmpdir):
        """Makes cwd point to the test's tmpdir."""
        with tmpdir.as_cwd():
            yield

    def test_init(self, env_name):
        """Tests default initialization given standard env specs."""
        gym.make(env_name)

    def test_reset(self, env_name):
        """Tests default initialization given standard env specs."""
        env = gym.make(env_name)
        obs = env.reset()
        assert env.observation_space.contains(obs)


@pytest.mark.parametrize("env_name", get_nethack_env_ids())
@pytest.mark.parametrize("rollout_len", [100])
class TestGymEnvRollout:
    @pytest.yield_fixture(autouse=True)  # will be applied to all tests in class
    def make_cwd_tmp(self, tmpdir):
        """Makes cwd point to the test's tmpdir."""
        with tmpdir.as_cwd():
            yield

    def test_rollout(self, env_name, rollout_len):
        """Tests rollout_len steps (or until termination) of random policy."""
        env = gym.make(env_name)
        rollout_env(env, rollout_len)

    def test_rollout_no_archive(self, env_name, rollout_len):
        """Tests rollout_len steps (or until termination) of random policy."""
        env = gym.make(env_name, archivefile=None)
        assert env.savedir is None
        assert env.archivefile is None
        assert env._stats_file is None
        assert env._stats_logger is None
        rollout_env(env, rollout_len)

    def test_seed_interface_output(self, env_name, rollout_len):
        """Tests whether env.seed output can be reused correctly."""
        env0 = gym.make(env_name)
        env1 = gym.make(env_name)

        seed_list0 = env0.seed()
        env0.reset()

        seed_dict = nle.env.seed_list_to_dict(seed_list0)
        assert env0.get_seeds() == seed_dict

        seed_list1 = env1.seed(seed_dict)
        assert seed_list0 == seed_list1

    def test_seed_rollout_from_nethack(self, env_name, rollout_len):
        """Tests that two NetHack instances with same seeds return same obs."""

        env0 = gym.make(env_name)
        env1 = gym.make(env_name)

        obs0 = env0.reset()  # no env.seed() call, so uses NetHack's seeds
        seeds = env0.get_seeds()

        env1.seed(seeds)
        obs1 = env1.reset()

        del obs0["message"]  # because of different names
        del obs1["message"]
        np.testing.assert_equal(obs0, obs1)
        compare_rollouts(env0, env1, rollout_len)

    def test_seed_rollout_seeded(self, env_name, rollout_len):
        """Tests that two seeded envs return same step data."""
        env0 = gym.make(env_name)
        env1 = gym.make(env_name)

        env0.seed()
        obs0 = env0.reset()
        seeds0 = env0.get_seeds()

        env1.seed(seeds0)
        obs1 = env1.reset()
        seeds1 = env1.get_seeds()

        assert seeds0 == seeds1

        del obs0["message"]  # because of different names
        del obs1["message"]
        np.testing.assert_equal(obs0, obs1)
        compare_rollouts(env0, env1, rollout_len)

    def test_seed_rollout_seeded_int(self, env_name, rollout_len):
        """Tests that two seeded envs return same step data."""
        env0 = gym.make(env_name)
        env1 = gym.make(env_name)

        env0.seed(random.randrange(sys.maxsize))
        obs0 = env0.reset()
        seeds0 = env0.get_seeds()

        env1.seed(seeds0)
        obs1 = env1.reset()
        seeds1 = env1.get_seeds()

        assert seeds0 == seeds1

        del obs0["message"]  # because of different names
        del obs1["message"]
        np.testing.assert_equal(obs0, obs1)
        compare_rollouts(env0, env1, rollout_len)

    def test_render_ansi(self, env_name, rollout_len):
        env = gym.make(env_name)
        env.reset()
        for _ in range(rollout_len):
            action = env.action_space.sample()
            _, _, done, _ = env.step(action)
            if done:
                env.reset()
            output = env.render(mode="ansi")
            assert isinstance(output, str)
            assert len(output.replace("\n", "")) == np.prod(nle.env.DUNGEON_SHAPE)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Status(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsStatus(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Status()
        x.Init(buf, n + offset)
        return x

    # Status
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Status
    def TITLE(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def STR(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def DX(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def CO(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def IN(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def WI(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def CH(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def ALIGN(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def SCORE(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def CAP(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def GOLD(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def ENE(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def ENEMAX(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def XP(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def AC(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def HD(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def TIME(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def HUNGER(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def HP(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def HPMAX(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def LEVELDESC(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(44))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def EXP(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(46))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def Condition(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(48))
        if o != 0:
            x = o + self._tab.Pos
            from nle.fbs.Condition import Condition
            obj = Condition()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def StatusStart(builder): builder.StartObject(23)
def StatusAddTITLE(builder, TITLE): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(TITLE), 0)
def StatusAddSTR(builder, STR): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(STR), 0)
def StatusAddDX(builder, DX): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(DX), 0)
def StatusAddCO(builder, CO): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(CO), 0)
def StatusAddIN(builder, IN): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(IN), 0)
def StatusAddWI(builder, WI): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(WI), 0)
def StatusAddCH(builder, CH): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(CH), 0)
def StatusAddALIGN(builder, ALIGN): builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(ALIGN), 0)
def StatusAddSCORE(builder, SCORE): builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(SCORE), 0)
def StatusAddCAP(builder, CAP): builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(CAP), 0)
def StatusAddGOLD(builder, GOLD): builder.PrependUOffsetTRelativeSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(GOLD), 0)
def StatusAddENE(builder, ENE): builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(ENE), 0)
def StatusAddENEMAX(builder, ENEMAX): builder.PrependUOffsetTRelativeSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(ENEMAX), 0)
def StatusAddXP(builder, XP): builder.PrependUOffsetTRelativeSlot(13, flatbuffers.number_types.UOffsetTFlags.py_type(XP), 0)
def StatusAddAC(builder, AC): builder.PrependUOffsetTRelativeSlot(14, flatbuffers.number_types.UOffsetTFlags.py_type(AC), 0)
def StatusAddHD(builder, HD): builder.PrependUOffsetTRelativeSlot(15, flatbuffers.number_types.UOffsetTFlags.py_type(HD), 0)
def StatusAddTIME(builder, TIME): builder.PrependUOffsetTRelativeSlot(16, flatbuffers.number_types.UOffsetTFlags.py_type(TIME), 0)
def StatusAddHUNGER(builder, HUNGER): builder.PrependUOffsetTRelativeSlot(17, flatbuffers.number_types.UOffsetTFlags.py_type(HUNGER), 0)
def StatusAddHP(builder, HP): builder.PrependUOffsetTRelativeSlot(18, flatbuffers.number_types.UOffsetTFlags.py_type(HP), 0)
def StatusAddHPMAX(builder, HPMAX): builder.PrependUOffsetTRelativeSlot(19, flatbuffers.number_types.UOffsetTFlags.py_type(HPMAX), 0)
def StatusAddLEVELDESC(builder, LEVELDESC): builder.PrependUOffsetTRelativeSlot(20, flatbuffers.number_types.UOffsetTFlags.py_type(LEVELDESC), 0)
def StatusAddEXP(builder, EXP): builder.PrependUOffsetTRelativeSlot(21, flatbuffers.number_types.UOffsetTFlags.py_type(EXP), 0)
def StatusAddCondition(builder, condition): builder.PrependStructSlot(22, flatbuffers.number_types.UOffsetTFlags.py_type(condition), 0)
def StatusEnd(builder): return builder.EndObject()



# Copyright (c) Facebook, Inc. and its affiliates.
from gym.envs import registration

from nle.env.base import NLE, DUNGEON_SHAPE, seed_list_to_dict

registration.register(id="NetHack-v0", entry_point="nle.env.base:NLE")

registration.register(id="NetHackScore-v0", entry_point="nle.env.tasks:NetHackScore")
registration.register(
    id="NetHackStaircase-v0", entry_point="nle.env.tasks:NetHackStaircase"
)
registration.register(
    id="NetHackStaircasePet-v0", entry_point="nle.env.tasks:NetHackStaircasePet"
)
registration.register(id="NetHackOracle-v0", entry_point="nle.env.tasks:NetHackOracle")
registration.register(id="NetHackGold-v0", entry_point="nle.env.tasks:NetHackGold")
registration.register(id="NetHackEat-v0", entry_point="nle.env.tasks:NetHackEat")
registration.register(id="NetHackScout-v0", entry_point="nle.env.tasks:NetHackScout")

__all__ = ["NLE", "DUNGEON_SHAPE"]



# This file taken from
#     https://github.com/deepmind/scalable_agent/blob/
#         cd66d00914d56c8ba2f0615d9cdeefcb169a8d70/vtrace.py
# and modified.

# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions to compute V-trace off-policy actor critic targets.

For details and theory see:

"IMPALA: Scalable Distributed Deep-RL with
Importance Weighted Actor-Learner Architectures"
by Espeholt, Soyer, Munos et al.

See https://arxiv.org/abs/1802.01561 for the full paper.
"""

import collections

import torch
import torch.nn.functional as F


VTraceFromLogitsReturns = collections.namedtuple(
    "VTraceFromLogitsReturns",
    [
        "vs",
        "pg_advantages",
        "log_rhos",
        "behavior_action_log_probs",
        "target_action_log_probs",
    ],
)

VTraceReturns = collections.namedtuple("VTraceReturns", "vs pg_advantages")


def action_log_probs(policy_logits, actions):
    return -F.nll_loss(
        F.log_softmax(torch.flatten(policy_logits, 0, -2), dim=-1),
        torch.flatten(actions),
        reduction="none",
    ).view_as(actions)


def from_logits(
    behavior_policy_logits,
    target_policy_logits,
    actions,
    discounts,
    rewards,
    values,
    bootstrap_value,
    clip_rho_threshold=1.0,
    clip_pg_rho_threshold=1.0,
):
    """V-trace for softmax policies."""

    target_action_log_probs = action_log_probs(target_policy_logits, actions)
    behavior_action_log_probs = action_log_probs(behavior_policy_logits, actions)
    log_rhos = target_action_log_probs - behavior_action_log_probs
    vtrace_returns = from_importance_weights(
        log_rhos=log_rhos,
        discounts=discounts,
        rewards=rewards,
        values=values,
        bootstrap_value=bootstrap_value,
        clip_rho_threshold=clip_rho_threshold,
        clip_pg_rho_threshold=clip_pg_rho_threshold,
    )
    return VTraceFromLogitsReturns(
        log_rhos=log_rhos,
        behavior_action_log_probs=behavior_action_log_probs,
        target_action_log_probs=target_action_log_probs,
        **vtrace_returns._asdict(),
    )


@torch.no_grad()
def from_importance_weights(
    log_rhos,
    discounts,
    rewards,
    values,
    bootstrap_value,
    clip_rho_threshold=1.0,
    clip_pg_rho_threshold=1.0,
):
    """V-trace from log importance weights."""
    with torch.no_grad():
        rhos = torch.exp(log_rhos)
        if clip_rho_threshold is not None:
            clipped_rhos = torch.clamp(rhos, max=clip_rho_threshold)
        else:
            clipped_rhos = rhos

        cs = torch.clamp(rhos, max=1.0)
        # Append bootstrapped value to get [v1, ..., v_t+1]
        values_t_plus_1 = torch.cat(
            [values[1:], torch.unsqueeze(bootstrap_value, 0)], dim=0
        )
        deltas = clipped_rhos * (rewards + discounts * values_t_plus_1 - values)

        acc = torch.zeros_like(bootstrap_value)
        result = []
        for t in range(discounts.shape[0] - 1, -1, -1):
            acc = deltas[t] + discounts[t] * cs[t] * acc
            result.append(acc)
        result.reverse()
        vs_minus_v_xs = torch.stack(result)

        # Add V(x_s) to get v_s.
        vs = torch.add(vs_minus_v_xs, values)

        # Advantage for policy gradient.
        broadcasted_bootstrap_values = torch.ones_like(vs[0]) * bootstrap_value
        vs_t_plus_1 = torch.cat(
            [vs[1:], broadcasted_bootstrap_values.unsqueeze(0)], dim=0
        )
        if clip_pg_rho_threshold is not None:
            clipped_pg_rhos = torch.clamp(rhos, max=clip_pg_rho_threshold)
        else:
            clipped_pg_rhos = rhos
        pg_advantages = clipped_pg_rhos * (rewards + discounts * vs_t_plus_1 - values)

        # Make sure no gradients backpropagated through the returned values.
        return VTraceReturns(vs=vs, pg_advantages=pg_advantages)



# Copyright (c) Facebook, Inc. and its affiliates.
import collections
import csv
import enum
import logging
import os
import random
import re
import sys
import time
import tempfile

import gym
import numpy as np

from nle import nethack
import nle.nethack.print_message as nhprint


logger = logging.getLogger(__name__)

WIN_MESSAGE = 1  # Technically dynamic. Practically constant.

FINAL_QUESTIONS = re.compile(
    rb"Do you want ("
    rb"your possessions identified|"
    rb"to see your attributes|"
    rb"an account of creatures vanquished|"
    rb"to see your conduct|"
    rb"to see the dungeon overview)"
)

DUNGEON_SHAPE = (21, 79)


def _fb_ndarray_to_np(fb_ndarray):
    result = fb_ndarray.DataAsNumpy()
    result = result.view(np.typeDict[fb_ndarray.Dtype()])
    result = result.reshape(fb_ndarray.ShapeAsNumpy().tolist())
    return result


INVFIELDS = [
    "Glyph",
    "Str",
    "Letter",
    "ObjectClass",
    # "ObjectClassName",
]

DEFAULT_MSG_PAD = 256
DEFAULT_INV_PAD = 55
DEFAULT_INVSTR_PAD = 80

ASCII_SPACE = ord(" ")
ASCII_y = ord("y")
ASCII_n = ord("n")
ASCII_ESC = nethack.C("[")

FULL_ACTIONS = list(nethack.ACTIONS)
# Removing some problematic actions
FULL_ACTIONS.remove(nethack.Command.SAVE)
# TODO: consider re-adding help eventually, when we can handle its state machine
# and output.
FULL_ACTIONS.remove(nethack.Command.HELP)
FULL_ACTIONS = tuple(FULL_ACTIONS)


# TODO(NN): this logic should be refactored into a class with pythonic access to
# its attributes.
def _get(response, path="Blstats.score", default=None, required=False):
    node = response

    for attr in path.split("."):
        attr = "".join(x.capitalize() for x in attr.split("_"))
        node = getattr(node, attr)()
        if node is None:
            if required:
                raise ValueError("%s not found in response, but required==True." % path)
            return default
    return node


def _get_glyphs(response):
    if response is None:
        return np.zeros(DUNGEON_SHAPE, dtype=np.int16)
    o = response.Observation()
    # If done is True, Observation() is None.
    if o is None:
        return np.zeros(DUNGEON_SHAPE, dtype=np.int16)
    return o.Glyphs().DataAsNumpy().view(np.int16).reshape(DUNGEON_SHAPE)


def _get_status_fast(response, entries=23):
    # Fast version of _get_status. See that function for the order.
    s = response.Blstats()
    if s is None:
        return np.zeros(entries, dtype=np.int32)
    return np.frombuffer(
        s._tab.Bytes[s._tab.Pos : s._tab.Pos + 4 * entries], dtype=np.int32
    )


def _get_status(response):
    s = response.Blstats()
    if s is None:
        return np.zeros(23, dtype=np.int32)

    return np.array(
        (
            s.CursX(),
            s.CursY(),
            # 1..125. See
            # https://nethackwiki.com/wiki/Attribute#Strength_in_game_formulas
            # and get_strength_str() in botl.c
            s.StrengthPercentage(),
            s.Strength(),
            s.Dexterity(),
            s.Constitution(),
            s.Intelligence(),
            s.Wisdom(),
            s.Charisma(),
            s.Score(),
            s.Hitpoints(),
            s.MaxHitpoints(),
            s.Depth(),
            s.Gold(),
            s.Energy(),
            s.MaxEnergy(),
            s.ArmorClass(),
            s.MonsterLevel(),
            s.ExperienceLevel(),
            s.ExperiencePoints(),
            s.Time(),
            s.HungerState(),
            s.CarryingCapacity(),
        ),
        dtype=np.int32,
    )


def _get_padded_message(
    response, padded_length=DEFAULT_MSG_PAD, alphabet_size=ord("~") - ord(" ") + 1
):
    # TODO: This would be faster if done in C++.
    result = np.full(padded_length, fill_value=alphabet_size, dtype=np.uint8)
    if response is None or response.NotRunning():
        return result

    win = response.Windows(WIN_MESSAGE)
    if win is None:
        logger.error("No message window. This shouldn't happen.")
        return result

    assert win.Type() == nethack.NHW_MESSAGE

    offset = 0
    for i in range(win.StringsLength()):
        message = np.frombuffer(win.Strings(i), dtype=np.uint8)

        # Subtract ord(" ") and assign. Crop if space runs out.
        result[offset : offset + len(message)] = (
            message[: max(len(result) - offset, 0)] - 0x20
        )
        offset += len(message) + 1  # Keep one separation token.
    return result


def _wait_for_space(response):
    internal = response.Internal()
    return internal and internal.Xwaitforspace()


def _in_moveloop(response):
    ps = response.ProgramState()
    return ps and ps.InMoveloop()


def _get_call_stack(response):
    internal = response.Internal()
    if internal is None:
        return ()
    return (internal.CallStack(i) for i in range(internal.CallStackLength()))


def _get_inv(response):
    result = {}
    for field in INVFIELDS:
        result[field] = []

    o = response.Observation()
    if o is None:
        return result

    for item in (o.Inventory(i) for i in range(o.InventoryLength())):
        for field in INVFIELDS:
            result[field].append(getattr(item, field)())
    return result


def _get_padded_inv(
    response,
    padded_length=DEFAULT_INV_PAD,
    str_padded_length=DEFAULT_INVSTR_PAD,
    alphabet_size=ord("~") - ord(" ") + 1,
):
    # TODO: This would be faster if done in C++.
    inv = _get_inv(response)
    strs = np.full(
        (padded_length, str_padded_length), fill_value=alphabet_size, dtype=np.uint8
    )
    for i, b in enumerate(inv["Str"]):
        strs[i, : len(b)] = np.frombuffer(b, dtype=np.uint8)[:str_padded_length] - 0x20

    pad_width = (0, padded_length - len(inv["Str"]))
    glyphs = np.pad(
        np.asarray(inv["Glyph"], dtype=np.int16),
        pad_width,
        mode="constant",
        constant_values=nethack.NO_GLYPH,
    )
    letters = np.pad(
        np.asarray(inv["Letter"], dtype=np.uint8) - 0x20,
        pad_width,
        mode="constant",
        constant_values=alphabet_size,
    )
    oclasses = np.pad(
        np.asarray(inv["ObjectClass"], dtype=np.uint8),
        pad_width,
        mode="constant",
        constant_values=nethack.MAXOCLASSES,
    )

    return glyphs, strs, letters, oclasses


class NLE(gym.Env):
    """Standard NetHack Learning Environment.

    Implements a gym interface around `nethack.NetHack`.


    Examples:
        >>> env = NLE()
        >>> obs = env.reset()
        >>> obs, reward, done, info = env.step(0)
        >>> env.render()
    """

    metadata = {"render.modes": ["human", "ansi"]}

    class StepStatus(enum.IntEnum):
        """Specifies the status of the terminal state.

        Note:
            User may redefine this class in subtasks to handle / categorize
            more terminal states.

            It is highly advised that, in such cases, 0 and 1 remain defined
            respectively as RUNNING and DEATH; redefining their semantics may
            break some logic elsewhere in the base environment. See
            `nle.env.tasks` for examples on how to do this right.
        """

        ABORTED = -1
        RUNNING = 0
        DEATH = 1

    Stats = collections.namedtuple(
        "Stats",
        (
            "end_status",
            "score",
            "time",
            "steps",
            "hp",
            "exp",
            "exp_lev",
            "gold",
            "hunger",
            "killer_name",
            "deepest_lev",
            "episode",
            "seeds",
            "ttyrec",
        ),
    )

    def __init__(
        self,
        savedir=None,
        archivefile="nethack.%(pid)i.%(time)s.zip",
        character="mon-hum-neu-mal",
        max_episode_steps=5000,
        observation_keys=("glyphs", "status", "message", "inventory"),
        actions=None,
        options=None,
    ):
        """Constructs a new NLE environment.

        Args:
            savedir (str or None): path to save archives into. Defaults to None.
            archivefile (str or None): Template for the zip archive filename of
                NetHack ttyrec files. Use "%(pid)i" for the process id of the
                NetHack process, "%(time)s" for the creation time. Use None to
                disable writing archivefiles.
            character (str): name of character. Defaults to "mon-hum-neu-mal".
            max_episode_steps (int): maximum amount of steps allowed before the
                game is forcefully quit. In such cases, ``info["end_status"]``
                will be equal to ``StepStatus.ABORTED``. Defaults to 5000.
            observation_keys (list): keys to use when creating the observation.
                Defaults to all.
            actions (list): list of actions. If None, the full action space will
                be used, i.e. ``nle.nethack.ACTIONS``. Defaults to None.
            options (list): list of game options to initialize NetHack. If None,
                NetHack will be initialized with the options found in
                ``nle.nethack.NETHACKOPTIONS`. Defaults to None.
        """

        self.character = character
        self._max_episode_steps = max_episode_steps

        if actions is None:
            actions = FULL_ACTIONS
        self._actions = actions

        self.response = None

        if archivefile is not None:
            try:
                if savedir is None:
                    parent_dir = os.path.join(os.getcwd(), "nle_data")
                    os.makedirs(parent_dir, exist_ok=True)
                    # Create a unique subdirectory for us.
                    self.savedir = tempfile.mkdtemp(
                        prefix=time.strftime("%Y%m%d-%H%M%S_"), dir=parent_dir
                    )
                else:
                    self.savedir = savedir
                    os.makedirs(self.savedir)
            except FileExistsError:
                logger.info("Using existing savedir: %s", self.savedir)
            else:
                logger.info("Created savedir: %s", self.savedir)

            self.archivefile = os.path.join(self.savedir, archivefile)
        else:
            self.savedir = None
            self.archivefile = None
            self._stats_file = None
            self._stats_logger = None

        self._setup_statsfile = archivefile is not None

        self.env = nethack.NetHack(
            archivefile=self.archivefile,
            options=options,
            playername="Agent%(pid)i-" + self.character,
        )

        self._random = random.SystemRandom()

        # -1 so that it's 0-based on first reset
        self._episode = -1

        space_dict = {
            "glyphs": gym.spaces.Box(
                low=np.iinfo(np.int16).min,
                high=np.iinfo(np.int16).max,
                shape=DUNGEON_SHAPE,
                dtype=np.int16,
            ),
            "status": gym.spaces.Box(
                low=np.iinfo(np.int32).min,
                high=np.iinfo(np.int32).max,
                shape=(23,),
                dtype=np.int32,
            ),
            "message": gym.spaces.Box(
                low=np.iinfo(np.uint8).min,
                high=np.iinfo(np.uint8).max,
                shape=(DEFAULT_MSG_PAD,),
                dtype=np.uint8,
            ),
            "inventory": gym.spaces.Tuple(
                (
                    gym.spaces.Box(
                        low=np.iinfo(np.int16).min,
                        high=np.iinfo(np.int16).max,
                        shape=(DEFAULT_INV_PAD,),
                        dtype=np.int16,
                    ),
                    gym.spaces.Box(
                        low=np.iinfo(np.uint8).min,
                        high=np.iinfo(np.uint8).max,
                        shape=(DEFAULT_INV_PAD, DEFAULT_INVSTR_PAD),
                        dtype=np.uint8,
                    ),
                    gym.spaces.Box(
                        low=np.iinfo(np.uint8).min,
                        high=np.iinfo(np.uint8).max,
                        shape=(DEFAULT_INV_PAD,),
                        dtype=np.uint8,
                    ),
                    gym.spaces.Box(
                        low=np.iinfo(np.uint8).min,
                        high=np.iinfo(np.uint8).max,
                        shape=(DEFAULT_INV_PAD,),
                        dtype=np.uint8,
                    ),
                )
            ),
        }

        self.observation_space = gym.spaces.Dict(
            {key: space_dict[key] for key in observation_keys}
        )

        self._key_functions = {
            "glyphs": _get_glyphs,
            "status": _get_status_fast,
            "message": _get_padded_message,
            "inventory": _get_padded_inv,
        }
        for key in list(self._key_functions.keys()):
            if key not in observation_keys:
                del self._key_functions[key]

        self.action_space = gym.spaces.Discrete(len(self._actions))

    def _get_observation(self, response):
        return {key: f(response) for key, f in self._key_functions.items()}

    def step(self, action: int):
        """Steps the environment.

        Args:
            action (int): action integer as defined by ``self.action_space``.

        Returns:
            (dict, float, bool, dict): a tuple containing
                - (*dict*): an observation of the state; this will contain the keys
                  specified by ``self.observation_space``.
                - (*float*): a reward; see ``self._reward_fn`` to see how it is
                  specified.
                - (*bool*): True if the state is terminal, False otherwise.
                - (*dict*): a dictionary of extra information (such as
                  `end_status`, i.e. a status info -- death, task win, etc. --
                  for the terminal state).
        """
        response, done, info = self.env.step(self._actions[action])
        response, done, info = self._perform_known_steps(response, done, info)

        self._steps += 1

        last_response = self.response
        self.response = response

        end_status = self.StepStatus(done or self._is_episode_end(response))
        reward = float(self._reward_fn(last_response, response, end_status))

        if end_status and not done:
            # Try to end the game nicely.
            self._quit_game(response, done, info)
            done = True

        if end_status:
            stats = self._collect_stats(last_response, end_status)
            stats = stats._asdict()
            info["stats"] = stats

            if self._stats_logger is not None:
                self._stats_logger.writerow(stats)

        info["end_status"] = end_status

        return self._get_observation(response), reward, done, info

    def _collect_stats(self, message, end_status):
        """Updates a stats dict tracking several env stats."""
        # Using class rather than instance to allow tasks to reuse this with
        # super()
        return NLE.Stats(
            end_status=int(end_status),
            score=_get(message, "Blstats.score", required=True),
            time=_get(message, "Blstats.time", required=True),
            steps=self._steps,
            hp=_get(message, "Blstats.hitpoints", required=True),
            exp=_get(message, "Blstats.experience_points", required=True),
            exp_lev=_get(message, "Blstats.experience_level", required=True),
            gold=_get(message, "Blstats.gold", required=True),
            hunger=_get(message, "You.uhunger", required=True),
            killer_name=self._killer_name,
            deepest_lev=_get(message, "Internal.deepest_lev_reached", required=True),
            episode=self._episode,
            seeds=self.get_seeds(),
            ttyrec=self.env._process.filename,
        )

    def reset(self):
        """Resets the environment.

        Note:
            We attempt to manually navigate the first few menus so that the
            first seen state is ready to be acted upon by the user. This might
            fail in case NetHack is initialized with some uncommon options.

        Returns:
            (dict): observation of the state as defined by
                    ``self.observation_space``
        """
        self.response = self.env.reset()

        # Only run on the first reset to initialize stats file
        if self._setup_statsfile:
            stats_file = os.path.splitext(self.env._archive.filename)[0] + ".csv"
            add_header = not os.path.exists(stats_file)

            self._stats_file = open(stats_file, "a", 1)  # line buffered
            self._stats_logger = csv.DictWriter(
                self._stats_file, fieldnames=self.Stats._fields
            )
            if add_header:
                self._stats_logger.writeheader()
        self._setup_statsfile = False

        self._episode += 1
        self._killer_name = "UNK"

        self._steps = 0
        self._info = self.env._info.copy()

        self._info["seeds"] = {}
        for k in nethack.SEED_KEYS:
            self._info["seeds"][k] = getattr(self.response.Seeds(), k.capitalize())()

        while not _in_moveloop(self.response):
            # Get past initial phase of game. This should make sure
            # all the observations are present.
            self.response, done, _ = self.env.step(ASCII_SPACE)
            assert not done, "Game ended unexpectedly"

        return self._get_observation(self.response)

    def get_seeds(self):
        """Returns current seeds.

        Returns:
            (dict): seeds used by the current instance of NetHack.
        """

        return self._info["seeds"].copy()

    def seed(self, seeds=None):
        """Seeds the environment.

        Won't take effect until the environment is reset. The game is seeded
        using the provided seed(s) each time that the environment is reset. Thus
        the environment will return one and only one instance of NetHack per
        seed.

        Arguments:
            seeds (dict): seed used on NetHack process upon reset. If type is
                          ``int``, a list of seeds gets generated by
                          incrementing the provided number. If ``None``, seeds
                          are generayed by NLE using ``self._random``.

        Returns:
            (list): the seeds used by NetHack (in a list to respect the API
                    defined by ``gym``). Use ``nle.seed_list_to_dict`` to get the
                    respective dictionary.
        """
        if seeds is None:
            seeds = {}
            for k in nethack.SEED_KEYS:
                seeds[k] = self._random.randrange(sys.maxsize)
        elif isinstance(seeds, int):
            to_add = 0
            seed_dict = {}
            for k in nethack.SEED_KEYS:
                to_add += 1
                seed_dict[k] = seeds + to_add
            seeds = seed_dict

        self.env.seed(seeds)
        return list(seeds.values())

    def render(self, mode="human"):
        """Renders the state of environment.

        Arguments:
           mode (str): Defaults to "human". Acceptable values are "human" and
                       "ansi", otherwise a ``ValueError`` is raised.

        """
        if mode == "human":
            nhprint.print_message(self.response)
            return
        elif mode == "ansi":
            # TODO(NN): refactor print_message and output string here
            chars = _get(self.response, "Observation.Chars")
            if chars is None:
                return ""
            chars = _fb_ndarray_to_np(chars)
            return "\n".join([line.tobytes().decode("utf-8") for line in chars])
        else:
            return super().render(mode=mode)

    def __repr__(self):
        return "<%s>" % self.__class__.__name__

    def _is_episode_end(self, response):
        """Returns whether the episode has ended.

        Tasks may override this method to specify different conditions, so long
        as the return value has a well defined __int__ method (e.g. booleans,
        numerical types, enum.IntEnum) and that value is part of StepStatus.

        The return value will be stored into info["end_status"].
        """
        if self._steps >= self._max_episode_steps:
            return self.StepStatus.ABORTED
        return self.StepStatus.RUNNING

    def _reward_fn(self, last_response, response, end_status):
        """Reward function. Difference between previous score and new score."""
        old_score = _get(last_response, "Blstats.score", 0)
        score = _get(response, "Blstats.score", old_score)
        del end_status  # Unused for "score" reward.
        return score - old_score

    def _perform_known_steps(self, response, done, info):
        while not done:
            if _wait_for_space(response):
                response, done, info = self.env.step(ASCII_SPACE)
                continue

            if self._killer_name == "UNK" and response.ProgramState().Gameover():
                self._killer_name = _get(
                    response, "Internal.killer_name", default="UNK"
                )
                if self._killer_name != "UNK":
                    self._killer_name = self._killer_name.decode("utf-8")

            message_win = response.Windows(WIN_MESSAGE)
            if message_win is None or not message_win.StringsLength():
                break
            msg = message_win.Strings(message_win.StringsLength() - 1)

            if msg.startswith(b"Beware, there will be no return!  Still climb?"):
                response, done, info = self.env.step(ASCII_n)
            elif re.match(FINAL_QUESTIONS, msg):
                response, done, info = self.env.step(ASCII_y)
            else:
                call_stack = _get_call_stack(response)
                if b"yn_function" in call_stack or b"getlin" in call_stack:
                    if b"eat" in msg or b"attack" in msg:
                        break
                    response, done, info = self.env.step(ASCII_ESC)
                else:
                    break

        return response, done, info

    def _quit_game(self, response, done, info):
        """Smoothly quit a game."""
        # Get out of menus and windows.
        response, done, info = self._perform_known_steps(response, done, info)

        if done:
            return

        # Quit the game.
        actions = "#quit\ny"
        for a in actions:
            response, done, info = self.env.step(ord(a))

        # Answer final questions.
        response, done, info = self._perform_known_steps(response, done, info)

        if not done:
            # Somehow, the above logic failed us. We'll SIGTERM the game to close it.
            if self.env._archive is None:
                filename = "N/A"
            else:
                filename = self.env._archive.filename
            logger.error(
                "Warning: smooth quitting of game failed, aborting "
                "(archive %s, episode %i).",
                filename,
                self.env._episode,
            )


def seed_list_to_dict(seeds):
    """Produces seeds dict out of the list of seeds returned by ``env.seed``.

    Arguments:
        seeds (list): list of seeds returned by ``NLE.seed``.

    Returns:
        (dict): seed dictionary correspondent to all seedable env vars.
    """
    return {k: seeds[i] for i, k in enumerate(nethack.SEED_KEYS)}



# Copyright (c) Facebook, Inc. and its affiliates.
import random
import sys
import time
import timeit
import multiprocessing as mp
import traceback


from nle import nethack


ACTIONS = [nethack.MiscAction.MORE]
ACTIONS += list(nethack.CompassDirection)
ACTIONS += list(nethack.CompassDirectionLonger)


def target(i, should_stop, queue):
    print("Starting", i)
    try:
        play(should_stop, queue)
    except KeyboardInterrupt:
        pass  # Return silently.
    except Exception as e:
        traceback.print_exc()
        print()
        raise e


def play(should_stop, queue):
    game = nethack.NetHack(archivefile=None)

    done = True
    steps = 0

    while not should_stop.is_set():
        if done or steps >= 1000:
            queue.put(steps)
            steps = 0
            observation = game.reset()

        if observation.Internal() and observation.Internal().Xwaitforspace():
            ch = nethack.MiscAction.MORE
        else:
            ch = random.choice(ACTIONS)

        observation, done, _ = game.step(ch)
        steps += 1


def main():
    num_games = int(sys.argv[1]) if len(sys.argv) > 1 else 10

    ctx = mp.get_context("fork")
    queue = ctx.Queue()
    should_stop = ctx.Event()

    steps = 0

    processes = []
    for i in range(num_games):
        p = ctx.Process(target=target, args=(i, should_stop, queue))
        p.start()
        processes.append(p)

    try:
        while steps < 1000000:
            start_time = timeit.default_timer()
            start_steps = steps
            time.sleep(5)
            while not queue.empty():
                steps += queue.get()
            end_time = timeit.default_timer()

            print((steps - start_steps) / (end_time - start_time))

        should_stop.set()
        for p in processes:
            p.join()

    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



# Copyright (c) Facebook, Inc. and its affiliates.
import functools
import logging
import os
import shutil
import tempfile
import time
import warnings
import weakref
import zipfile

from . import ptyprocess
import zmq


with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    # Import all flatbuffer modules.
    from nle.fbs import (  # noqa: F401
        Blstats,
        Condition,
        DLevel,
        Internal,
        InventoryItem,
        MenuItem,
        Message,
        NDArray,
        Observation,
        ProgramState,
        Seeds,
        Status,
        Window,
        You,
    )


SEED_KEYS = ["core", "disp"]

NETHACKOPTIONS = [
    "windowtype:rl",
    "color",
    "showexp",
    "autopickup",
    "pickup_types:$?!/",
    "pickup_burden:unencumbered",
]

HACKDIR = os.getenv("HACKDIR")

if HACKDIR is None:
    # Somewhat HACKy way of getting HACKDIR from installed nethack.
    script = shutil.which("nethack")
    if script is None:
        raise FileNotFoundError("Didn't find nethack in path. Is it installed?")
    with open(script, "r") as f:
        while True:
            line = f.readline()
            if not line:
                raise ValueError("Could not determine HACKDIR from installed nethack")
            if line.startswith("HACKDIR="):
                HACKDIR = line[8:-1]  # Exclude newline.
                break


EXECUTABLE = os.path.join(HACKDIR, "nethack")
if not os.path.exists(EXECUTABLE):
    raise FileNotFoundError(
        "Couldn't run nethack in %s as file doesn't exist" % EXECUTABLE
    )


def _exec_nethack(playername, hackdir, seeds=None, options=NETHACKOPTIONS):
    """Turns current process into NetHack with right environment variables."""
    user = playername % {"pid": os.getpid()}

    env = {
        "HACKDIR": hackdir,
        "NETHACKOPTIONS": ",".join(options),
        "USER": user,
        "SHELL": "/bin/bash",
        "TERM": "xterm-256color",
    }

    if seeds is not None:
        for name, seed in seeds.items():
            name = "NLE_SEED_" + name.upper()
            env[name] = str(seed)

    command = EXECUTABLE + " -u" + user

    shell = os.environ.get("SHELL", "/bin/bash")
    os.execle(shell, os.path.basename(shell), "-c", command, env)
    raise OSError("This shouldn't happen.")


def _finalize_one_run(hackdir):
    prefix = str(os.getuid())
    if not os.path.exists(hackdir):
        # Removed in finalizer of NetHack object. Also fine.
        return

    with os.scandir(hackdir) as scandir:
        for entry in scandir:
            if entry.name.startswith(prefix):
                os.unlink(entry.path)


def _recordclosefn(archive, filename):
    archive.write(filename)
    os.unlink(filename)


class NetHack:
    def __init__(
        self,
        archivefile="nethack.%(pid)i.%(time)s.zip",
        playername="Agent%(pid)i-mon-hum-neu-mal",
        options=None,
        rows=24,
        columns=80,
        context=None,
    ):
        """Constructs a new NetHack environment."""
        self._playername = playername
        self._rows = rows
        self._columns = columns

        if options is None:
            options = NETHACKOPTIONS
        self._nethackoptions = options

        self._episode = 0
        self._info = {}

        self._finalizers = []

        if not os.path.exists(HACKDIR) or not os.path.exists(
            os.path.join(HACKDIR, "nethack")
        ):
            raise FileNotFoundError("Couldn't find NetHack installation.")

        # Create a HACKDIR for us.
        self._vardir = tempfile.mkdtemp(prefix="nle")

        os.symlink(os.path.join(HACKDIR, "nhdat"), os.path.join(self._vardir, "nhdat"))

        # touch a few files.
        for filename in ["sysconf", "perm", "logfile", "xlogfile"]:
            os.close(os.open(os.path.join(self._vardir, filename), os.O_CREAT))
        os.mkdir(os.path.join(self._vardir, "save"))

        if archivefile is None:
            self._archive = None
            self._recordclosefn = lambda f: None
        else:
            try:
                self._archive = zipfile.ZipFile(
                    archivefile
                    % {"pid": os.getpid(), "time": time.strftime("%Y%m%d-%H%M%S")},
                    "x",
                )
            except FileExistsError:
                logging.exception("Archive file %s exists, terminating" % archivefile)
                raise

            self._recordclosefn = functools.partial(_recordclosefn, self._archive)
            self._finalizers.append(
                # Cannot close archive before final call to _recordclosefn.
                # Binding to lifetime of self doesn't work here, so bind to
                # _recordclosefn itself.
                weakref.finalize(self._recordclosefn, self._archive.close)
            )
            logging.info("Archiving replays in %s" % self._archive.filename)

        self._context = context or zmq.Context.instance()

        self._finalizers.append(weakref.finalize(self, shutil.rmtree, self._vardir))

        self._exec_nethack = None
        self.seed(None)  # Sets self._exec_nethack.

        # TODO(heiner): Consider having a collections.deque of processes for
        # pre-loading. This requires us to have different connections, which
        # would be great to support more than one NetHack object.
        self._process = None

    def _recv(self):
        buf = self._socket.recv()
        message = Message.Message.GetRootAsMessage(buf, 0)
        # TODO(heiner): Consider waitpid'ing to get process status.
        return message, message.NotRunning()

    def reset(self):
        if self._archive is None:
            self.recordname = None
        else:
            self.recordname = "nethack.run.%i.%%(time)s.%%(pid)i.ttyrec" % self._episode
        self._process = ptyprocess.PtyProcess(
            target=self._exec_nethack,
            recordclosefn=self._recordclosefn,
            recordname=self.recordname,
        )
        self._process.fork(rows=self._rows, columns=self._columns, wait_for_output=True)

        socketfile = os.path.join(self._vardir, "%i.nle.sock" % self._process.pid)
        address = "ipc://" + socketfile
        logging.debug("Connecting to %s...", address)
        self._socket = self._context.socket(zmq.PULL)
        self._socket.connect(address)

        weakref.finalize(self._process, self._socket.close)
        weakref.finalize(self._process, _finalize_one_run, self._vardir)

        if not self._socket.poll(timeout=1000):  # 1s timeout.
            raise IOError("No response received from NetHack process -- is it running?")

        message, done = self._recv()
        assert not done, "NetHack closed without input."

        # Connection established, can remove socket file from file system.
        os.unlink(socketfile)

        self._info["pid"] = self._process.pid
        self._info["episode"] = self._episode

        self._episode += 1

        return message

    def step(self, action):
        self._process.write(bytes((action,)))
        message, done = self._recv()

        return message, done, self._info

    def close(self):
        del self._process  # Triggers finalizer.
        for f in self._finalizers:
            f()

    def seed(self, seeds):
        if isinstance(seeds, dict):
            for k in SEED_KEYS:
                if k not in seeds:
                    continue
                if seeds[k] < 1:
                    raise ValueError(
                        "Found %d for %s, but seeds must be positive integers.",
                        seeds[k],
                        k,
                    )
        elif seeds is not None:
            raise ValueError("`seeds` is %s, but must be either None or a dict.", seeds)

        self._seeds = seeds

        self._exec_nethack = functools.partial(
            _exec_nethack,
            self._playername,
            self._vardir,
            self._seeds,
            self._nethackoptions,
        )



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import gnuplotlib as gp
import numpy as np
import pandas as pd

from pathlib import Path


def str_to_float_pair(s):
    """
    Convert string to pair of floats.
    """
    if s is None:
        return None
    split = s.split(",")
    if len(split) != 2:
        raise RuntimeError("range does not match pattern 'float,float'")
    return (float(split[0]), float(split[1]))


parser = argparse.ArgumentParser("NetHack GnuPlotter", allow_abbrev=False)
parser.register("type", "pair", str_to_float_pair)
parser.add_argument(
    "-f",
    "--file",
    type=str,
    default="~/torchbeast/latest/logs.tsv",
    help="file to plot or directory to look for log files",
)
parser.add_argument(
    "-w", "--window", type=int, default=-1, help="override automatic window size."
)
parser.add_argument("-x", "--width", type=int, default=80, help="width of plot")
parser.add_argument("-y", "--height", type=int, default=30, help="height of plot")
parser.add_argument(
    "--no_legend",
    action="store_true",
    help="skip printing legend when plotting multiple experiments",
)
parser.add_argument(
    "--xrange",
    type="pair",
    default=None,
    help="float,float. range of x values to plot. overrides automatic zoom for x axis.",
)
parser.add_argument(
    "--yrange",
    type="pair",
    default=None,
    help="float,float. range of y values to plot. overrides automatic zoom for y axis.",
)


def plot_single_ascii(target, width, height, window, xrange=None, yrange=None):
    """
    Plot the target file using the specified width and height.
    If window > 0, use it to specify the window size for rolling averages.
    """
    print("plotting %s" % str(target))
    df = pd.read_csv(target, sep="\t")
    steps = np.array(df["# Step"])

    if window < 0:
        window = len(steps) // width + 1
    window = df["mean_episode_return"].rolling(window=window, min_periods=0)
    returns = np.array(window.mean())
    stderrs = np.array(window.std())

    plot_options = {}
    plot_options["with"] = "yerrorbars"
    plot_options["terminal"] = "dumb %d %d ansi" % (width, height)
    plot_options["tuplesize"] = 3
    plot_options["title"] = "averaged episode return"
    plot_options["xlabel"] = "steps"

    if xrange is not None:
        plot_options["xrange"] = xrange

    if yrange is not None:
        plot_options["yrange"] = yrange

    gp.plot(steps, returns, stderrs, **plot_options)


def plot_multiple_ascii(
    target, width, height, window, xrange=None, yrange=None, no_legend=False
):
    """
    Plot files under the target path using the specified width and height.
    If window > 0, use it to specify the window size for rolling averages.
    Use xrange and yrange to override the autozoom to look at specific parts
    of the plot.

    Set no_legend to true to save the visual space for the plot.
    """
    dfs = []
    if target.is_dir():
        # fully specified directory
        glob = target.glob("**/logs.tsv")
    else:
        # partially specified directory, glob from parent dir
        glob = target.parent.glob("**/logs.tsv")
    for child in glob:
        if target.name in str(child.parent):
            # this allows users to provide partial dir names, e.g. "~/logs/2020-"
            try:
                df = pd.read_csv(child, sep="\t")
                # TODO: remove rows with nan? maybe bad as will damage rolling window
                # df[df["mean_episode_return"] == df["mean_episode_return"]].copy()
                if len(df) > 0:
                    name = "    " + str(child.parent) + ":"
                    dfs.append((name, df))
            except pd.errors.EmptyDataError:
                # nothing to plot
                pass

    if len(dfs) == 0:
        # didn't find any valid tsv logs
        raise FileNotFoundError("No logs found under %s" % target)

    if window < 0:
        max_size = max(len(df["# Step"]) for name, df in dfs)
        window = max_size // width + 1

    datasets = []
    for name, df in dfs:
        steps = np.array(df["# Step"])
        if window > 1:
            roll = df["mean_episode_return"].rolling(window=window, min_periods=0)
            try:
                rewards = np.array(roll.mean())
            except pd.core.base.DataError:
                print("Error reading file at %s" % name)
                continue
        else:
            rewards = np.array(df["mean_episode_return"])
        if no_legend:
            datasets.append((steps, rewards))
        else:
            datasets.append((steps, rewards, dict(legend=name)))

    errs = len(dfs) - len(datasets)
    if errs > 0:
        print(
            "Skipped %d runs (%f) due to errors reading data" % (errs, errs / len(dfs))
        )
    print(
        "Plotting %d runs with window_size %d from %s" % (len(datasets), window, target)
    )

    plot_options = {}
    plot_options["terminal"] = "dumb %d %d ansi" % (width, height)
    plot_options["tuplesize"] = 2
    plot_options["title"] = "averaged episode return"
    plot_options["xlabel"] = "steps"
    plot_options["set"] = "key below"

    if xrange is not None:
        plot_options["xrange"] = xrange

    if yrange is not None:
        plot_options["yrange"] = yrange

    gp.plot(*datasets, **plot_options)


def plot(flags):
    target = Path(flags.file).expanduser()

    if target.is_file():
        # plot single torchbeast run, path/to/logs.tsv
        if target.suffix == ".tsv":
            plot_single_ascii(
                target,
                flags.width,
                flags.height,
                flags.window,
                flags.xrange,
                flags.yrange,
            )
        else:
            raise RuntimeError(
                "Filetype not recognised (expected .tsv): %s" % target.suffix
            )
    elif (target / "logs.tsv").is_file():
        # next check if this is actually a single run directory with file "logs.tsv"
        plot_single_ascii(
            target / "logs.tsv",
            flags.width,
            flags.height,
            flags.window,
            flags.xrange,
            flags.yrange,
        )
    else:
        # look for runs underneath the specified directory
        plot_multiple_ascii(
            target,
            flags.width,
            flags.height,
            flags.window,
            flags.xrange,
            flags.yrange,
            flags.no_legend,
        )


if __name__ == "__main__":
    flags = parser.parse_args()
    plot(flags)



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
import argparse
import fcntl
import os
import signal
import struct
import termios
import time
import tty

parser = argparse.ArgumentParser()
parser.add_argument(
    "-1",
    "--no_input",
    action="store_true",
    help=("Save in ttyrec (not ttyrec2) format, " "without input data"),
)
parser.add_argument(
    "-e", "--execute", metavar="CMD", help="command to run (default: /bin/sh)"
)
parser.add_argument(
    "-a",
    "--append",
    action="store_true",
    help="append instead of truncating output file",
)
parser.add_argument("--keep_stderr", action="store_true", help="don't process stderr")
parser.add_argument(
    "filename", default="out.ttyrec", type=str, nargs="?", help="tty record file"
)


def write_header(fp, length, channel):
    try:
        usec = time.time_ns() // 1000
    except AttributeError:  # time_ns is only available in Python 3.7+.
        usec = int(1000 * 1000 * time.time())
    sec = usec // 1000 // 1000
    usec -= 1000 * 1000 * sec  # now = sec + usec * 1e-6

    if FLAGS.no_input:
        os.write(fp, struct.pack("<iii", sec, usec, length))
    else:
        os.write(fp, struct.pack("<iiiB", sec, usec, length, channel))


def dooutput(fdmaster, fdslave, script):
    os.close(0)
    os.close(fdslave)

    while True:
        buf = os.read(fdmaster, 1024)
        if not buf:
            break
        os.write(1, buf)
        fcntl.flock(script, fcntl.LOCK_EX)
        write_header(script, len(buf), 0)
        os.write(script, buf)
        fcntl.flock(script, fcntl.LOCK_UN)

    os.close(fdmaster)
    os._exit(0)


def doshell(fdmaster, fdslave):
    os.setsid()
    fcntl.ioctl(fdslave, termios.TIOCSCTTY)

    os.close(fdmaster)
    os.dup2(fdslave, 0)
    os.dup2(fdslave, 1)
    if not FLAGS.keep_stderr:
        os.dup2(fdslave, 2)
    os.close(fdslave)

    shell = os.environ.get("SHELL", "/bin/bash")

    if FLAGS.execute:
        os.execl(shell, os.path.basename(shell), "-c", FLAGS.execute)
    else:
        os.execl(shell, os.path.basename(shell))


def doinput(fdmaster, fdslave, script):
    while True:
        buf = os.read(0, 1024)
        if not buf:
            break
        os.write(fdmaster, buf)

        if FLAGS.no_input:
            continue

        fcntl.flock(script, fcntl.LOCK_EX)
        write_header(script, len(buf), 1)
        os.write(script, buf)
        fcntl.flock(script, fcntl.LOCK_UN)


def main():
    global FLAGS
    FLAGS = parser.parse_args()

    flags = os.O_CREAT | os.O_WRONLY
    if FLAGS.append:
        flags |= os.O_APPEND
    else:
        flags |= os.O_TRUNC
    scriptfd = os.open(FLAGS.filename, flags)
    os.close(scriptfd)

    winsz = bytearray(8)
    fcntl.ioctl(0, termios.TIOCGWINSZ, winsz)

    fdmaster, fdslave = os.openpty()
    fcntl.ioctl(fdslave, termios.TIOCSWINSZ, winsz)

    tt = termios.tcgetattr(0)

    tty.setraw(0)
    rtt = termios.tcgetattr(0)
    rtt[3] &= ~termios.ECHO  # lflags
    termios.tcsetattr(0, termios.TCSAFLUSH, rtt)

    scriptfd = 0
    child = 0
    subchild = 0

    def finish(signum, frame):
        pid, status = os.waitpid(-1, os.WNOHANG)
        if pid == child:
            termios.tcsetattr(0, termios.TCSAFLUSH, tt)

        os.close(scriptfd)
        os._exit(0)

    signal.signal(signal.SIGCHLD, finish)

    child = os.fork()

    if child == 0:
        subchild = os.fork()

        if subchild != 0:
            scriptfd = os.open(FLAGS.filename, os.O_APPEND | os.O_WRONLY)
            dooutput(fdmaster, fdslave, scriptfd)
        else:
            doshell(fdmaster, fdslave)
    try:
        scriptfd = os.open(FLAGS.filename, os.O_APPEND | os.O_WRONLY)
        doinput(fdmaster, fdslave, scriptfd)
    finally:
        termios.tcsetattr(0, termios.TCSAFLUSH, tt)


if __name__ == "__main__":
    main()



# Copyright (c) Facebook, Inc. and its affiliates.
import os
import unittest
import tempfile

import numpy as np

from nle import nethack


def _fb_ndarray_to_np(fb_ndarray):
    result = fb_ndarray.DataAsNumpy()
    result = result.view(np.typeDict[fb_ndarray.Dtype()])
    result = result.reshape(fb_ndarray.ShapeAsNumpy().tolist())
    return result


class NetHackTest(unittest.TestCase):
    def test_run(self):
        archivefile = tempfile.mktemp(suffix="nethack_test", prefix=".zip")
        game = nethack.NetHack(archivefile=archivefile)

        response = game.reset()
        actions = [
            nethack.MiscAction.MORE,
            nethack.MiscAction.MORE,
            nethack.MiscAction.MORE,
            nethack.MiscAction.MORE,
            nethack.MiscAction.MORE,
            nethack.MiscAction.MORE,
        ]

        for action in actions:
            while not response.ProgramState().InMoveloop():
                response, done, info = game.step(nethack.MiscAction.MORE)

            response, done, info = game.step(action)
            if done:
                # Only the good die young.
                response = game.reset()

            obs = response.Observation()
            chars = _fb_ndarray_to_np(obs.Chars())
            glyphs = _fb_ndarray_to_np(obs.Glyphs())

            status = response.Blstats()
            x, y = status.CursX(), status.CursY()

            self.assertEqual(np.count_nonzero(chars == ord("@")), 1)
            self.assertEqual(chars[y, x], ord("@"))

            mon = nethack.glyph_to_mon(glyphs[y][x])
            self.assertEqual(mon.mname, "monk")
            self.assertEqual(mon.mlevel, 10)

            class_sym = nethack.mlet_to_class_sym(mon.mlet)
            self.assertEqual(class_sym.sym, "@")
            self.assertEqual(class_sym.explain, "human or elf")

        self.assertEqual(os.waitpid(info["pid"], os.WNOHANG), (0, 0))

        del game  # Should kill process.

        with self.assertRaisesRegex(OSError, "No (child|such)? process"):
            os.waitpid(info["pid"], 0)


class HelperTest(unittest.TestCase):
    def test_simple(self):
        glyph = 155  # Lichen.

        mon = nethack.glyph_to_mon(glyph)

        self.assertEqual(mon.mname, "lichen")

        cs = nethack.mlet_to_class_sym(mon.mlet)

        self.assertEqual(cs.sym, "F")
        self.assertEqual(cs.explain, "fungus or mold")

        self.assertEqual(nethack.NHW_MESSAGE, 1)
        self.assertTrue(hasattr(nethack, "MAXWIN"))


if __name__ == "__main__":
    unittest.main()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Message(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsMessage(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Message()
        x.Init(buf, n + offset)
        return x

    # Message
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Message
    def Observation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from nle.fbs.Observation import Observation
            obj = Observation()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Message
    def Blstats(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from nle.fbs.Blstats import Blstats
            obj = Blstats()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Message
    def You(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from nle.fbs.You import You
            obj = You()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Message
    def Windows(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from nle.fbs.Window import Window
            obj = Window()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Message
    def WindowsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Message
    def WindowsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Message
    def Internal(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from nle.fbs.Internal import Internal
            obj = Internal()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Message
    def ProgramState(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = o + self._tab.Pos
            from nle.fbs.ProgramState import ProgramState
            obj = ProgramState()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Message
    def Seeds(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = o + self._tab.Pos
            from nle.fbs.Seeds import Seeds
            obj = Seeds()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Message
    def NotRunning(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def MessageStart(builder): builder.StartObject(8)
def MessageAddObservation(builder, observation): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(observation), 0)
def MessageAddBlstats(builder, blstats): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(blstats), 0)
def MessageAddYou(builder, you): builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(you), 0)
def MessageAddWindows(builder, windows): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(windows), 0)
def MessageStartWindowsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def MessageAddInternal(builder, internal): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(internal), 0)
def MessageAddProgramState(builder, programState): builder.PrependStructSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(programState), 0)
def MessageAddSeeds(builder, seeds): builder.PrependStructSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(seeds), 0)
def MessageAddNotRunning(builder, notRunning): builder.PrependBoolSlot(7, notRunning, 0)
def MessageEnd(builder): return builder.EndObject()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This is an example self-contained agent running NLE based on MonoBeast.

import argparse
import logging
import os
import pprint
import threading
import time
import timeit
import traceback

# Necessary for multithreading.
os.environ["OMP_NUM_THREADS"] = "1"  # noqa: E402

try:
    import torch
    from torch import multiprocessing as mp
    from torch import nn
    from torch.nn import functional as F
except ImportError:
    logging.exception(
        "PyTorch not found. Please install the agent dependencies with "
        '`pip install ".[agent]"`'
    )

import gym

import nle  # noqa: F401
from nle.agent import vtrace


# yapf: disable
parser = argparse.ArgumentParser(description="PyTorch Scalable Agent")

parser.add_argument("--env", type=str, default="NetHackScore-v0",
                    help="Gym environment.")
parser.add_argument("--mode", default="train",
                    choices=["train", "test", "test_render"],
                    help="Training or test mode.")

# Training settings.
parser.add_argument("--disable_checkpoint", action="store_true",
                    help="Disable saving checkpoint.")
parser.add_argument("--savedir", default="~/torchbeast/",
                    help="Root dir where experiment data will be saved.")
parser.add_argument("--num_actors", default=4, type=int, metavar="N",
                    help="Number of actors (default: 4).")
parser.add_argument("--total_steps", default=100000, type=int, metavar="T",
                    help="Total environment steps to train for.")
parser.add_argument("--batch_size", default=8, type=int, metavar="B",
                    help="Learner batch size.")
parser.add_argument("--unroll_length", default=80, type=int, metavar="T",
                    help="The unroll length (time dimension).")
parser.add_argument("--num_buffers", default=None, type=int,
                    metavar="N", help="Number of shared-memory buffers.")
parser.add_argument("--num_learner_threads", "--num_threads", default=2, type=int,
                    metavar="N", help="Number learner threads.")
parser.add_argument("--disable_cuda", action="store_true",
                    help="Disable CUDA.")
parser.add_argument("--use_lstm", action="store_true",
                    help="Use LSTM in agent model.")

# Loss settings.
parser.add_argument("--entropy_cost", default=0.0006,
                    type=float, help="Entropy cost/multiplier.")
parser.add_argument("--baseline_cost", default=0.5,
                    type=float, help="Baseline cost/multiplier.")
parser.add_argument("--discounting", default=0.99,
                    type=float, help="Discounting factor.")
parser.add_argument("--reward_clipping", default="abs_one",
                    choices=["abs_one", "none"],
                    help="Reward clipping.")

# Optimizer settings.
parser.add_argument("--learning_rate", default=0.00048,
                    type=float, metavar="LR", help="Learning rate.")
parser.add_argument("--alpha", default=0.99, type=float,
                    help="RMSProp smoothing constant.")
parser.add_argument("--momentum", default=0, type=float,
                    help="RMSProp momentum.")
parser.add_argument("--epsilon", default=0.01, type=float,
                    help="RMSProp epsilon.")
parser.add_argument("--grad_norm_clipping", default=40.0, type=float,
                    help="Global gradient norm clip.")
# yapf: enable


logging.basicConfig(
    format=(
        "[%(levelname)s:%(process)d %(module)s:%(lineno)d %(asctime)s] " "%(message)s"
    ),
    level=logging.INFO,
)


def nested_map(f, n):
    if isinstance(n, tuple) or isinstance(n, list):
        return n.__class__(nested_map(f, sn) for sn in n)
    elif isinstance(n, dict):
        return {k: nested_map(f, v) for k, v in n.items()}
    else:
        return f(n)


def compute_baseline_loss(advantages):
    return 0.5 * torch.sum(advantages ** 2)


def compute_entropy_loss(logits):
    """Return the entropy loss, i.e., the negative entropy of the policy."""
    policy = F.softmax(logits, dim=-1)
    log_policy = F.log_softmax(logits, dim=-1)
    return torch.sum(policy * log_policy)


def compute_policy_gradient_loss(logits, actions, advantages):
    cross_entropy = F.nll_loss(
        F.log_softmax(torch.flatten(logits, 0, 1), dim=-1),
        target=torch.flatten(actions, 0, 1),
        reduction="none",
    )
    cross_entropy = cross_entropy.view_as(advantages)
    return torch.sum(cross_entropy * advantages.detach())


def create_env(name, *args, **kwargs):
    return gym.make(name, observation_keys=("glyphs", "status"), *args, **kwargs)


def act(
    flags,
    actor_index: int,
    free_queue: mp.SimpleQueue,
    full_queue: mp.SimpleQueue,
    model: torch.nn.Module,
    buffers,
    initial_agent_state_buffers,
):
    try:
        logging.info("Actor %i started.", actor_index)

        gym_env = create_env(
            flags.env,
            savedir=flags.rundir,
            archivefile="nethack.%i.%%(pid)i.%%(time)s.zip" % actor_index,
        )
        env = ResettingEnvironment(gym_env)
        env_output = env.initial()
        agent_state = model.initial_state(batch_size=1)
        agent_output, unused_state = model(env_output, agent_state)
        while True:
            index = free_queue.get()
            if index is None:
                break

            # Write old rollout end.
            for key in env_output:
                buffers[key][index][0, ...] = env_output[key]
            for key in agent_output:
                buffers[key][index][0, ...] = agent_output[key]
            for i, tensor in enumerate(agent_state):
                initial_agent_state_buffers[index][i][...] = tensor

            # Do new rollout.
            for t in range(flags.unroll_length):
                with torch.no_grad():
                    agent_output, agent_state = model(env_output, agent_state)

                env_output = env.step(agent_output["action"])

                for key in env_output:
                    buffers[key][index][t + 1, ...] = env_output[key]
                for key in agent_output:
                    buffers[key][index][t + 1, ...] = agent_output[key]

            full_queue.put(index)

    except KeyboardInterrupt:
        pass  # Return silently.
    except Exception:
        logging.error("Exception in worker process %i", actor_index)
        traceback.print_exc()
        print()
        raise


def get_batch(
    flags,
    free_queue: mp.SimpleQueue,
    full_queue: mp.SimpleQueue,
    buffers,
    initial_agent_state_buffers,
    lock=threading.Lock(),
):
    with lock:
        indices = [full_queue.get() for _ in range(flags.batch_size)]
    batch = {
        key: torch.stack([buffers[key][m] for m in indices], dim=1) for key in buffers
    }
    initial_agent_state = (
        torch.cat(ts, dim=1)
        for ts in zip(*[initial_agent_state_buffers[m] for m in indices])
    )
    for m in indices:
        free_queue.put(m)
    batch = {k: t.to(device=flags.device, non_blocking=True) for k, t in batch.items()}
    initial_agent_state = tuple(
        t.to(device=flags.device, non_blocking=True) for t in initial_agent_state
    )
    return batch, initial_agent_state


def learn(
    flags,
    actor_model,
    model,
    batch,
    initial_agent_state,
    optimizer,
    scheduler,
    lock=threading.Lock(),  # noqa: B008
):
    """Performs a learning (optimization) step."""
    with lock:
        learner_outputs, unused_state = model(batch, initial_agent_state)

        # Take final value function slice for bootstrapping.
        bootstrap_value = learner_outputs["baseline"][-1]

        # Move from obs[t] -> action[t] to action[t] -> obs[t].
        batch = {key: tensor[1:] for key, tensor in batch.items()}
        learner_outputs = {key: tensor[:-1] for key, tensor in learner_outputs.items()}

        rewards = batch["reward"]
        if flags.reward_clipping == "abs_one":
            clipped_rewards = torch.clamp(rewards, -1, 1)
        elif flags.reward_clipping == "none":
            clipped_rewards = rewards

        discounts = (~batch["done"]).float() * flags.discounting

        vtrace_returns = vtrace.from_logits(
            behavior_policy_logits=batch["policy_logits"],
            target_policy_logits=learner_outputs["policy_logits"],
            actions=batch["action"],
            discounts=discounts,
            rewards=clipped_rewards,
            values=learner_outputs["baseline"],
            bootstrap_value=bootstrap_value,
        )

        pg_loss = compute_policy_gradient_loss(
            learner_outputs["policy_logits"],
            batch["action"],
            vtrace_returns.pg_advantages,
        )
        baseline_loss = flags.baseline_cost * compute_baseline_loss(
            vtrace_returns.vs - learner_outputs["baseline"]
        )
        entropy_loss = flags.entropy_cost * compute_entropy_loss(
            learner_outputs["policy_logits"]
        )

        total_loss = pg_loss + baseline_loss + entropy_loss

        episode_returns = batch["episode_return"][batch["done"]]
        stats = {
            "episode_returns": tuple(episode_returns.cpu().numpy()),
            "mean_episode_return": torch.mean(episode_returns).item(),
            "total_loss": total_loss.item(),
            "pg_loss": pg_loss.item(),
            "baseline_loss": baseline_loss.item(),
            "entropy_loss": entropy_loss.item(),
        }

        optimizer.zero_grad()
        total_loss.backward()
        nn.utils.clip_grad_norm_(model.parameters(), flags.grad_norm_clipping)
        optimizer.step()
        scheduler.step()

        actor_model.load_state_dict(model.state_dict())
        return stats


def create_buffers(flags, observation_space, num_actions, num_overlapping_steps=1):
    size = (flags.unroll_length + num_overlapping_steps,)

    # Get specimens to infer shapes and dtypes.
    samples = {k: torch.from_numpy(v) for k, v in observation_space.sample().items()}

    specs = {
        key: dict(size=size + sample.shape, dtype=sample.dtype)
        for key, sample in samples.items()
    }
    specs.update(
        reward=dict(size=size, dtype=torch.float32),
        done=dict(size=size, dtype=torch.bool),
        episode_return=dict(size=size, dtype=torch.float32),
        episode_step=dict(size=size, dtype=torch.int32),
        policy_logits=dict(size=size + (num_actions,), dtype=torch.float32),
        baseline=dict(size=size, dtype=torch.float32),
        last_action=dict(size=size, dtype=torch.int64),
        action=dict(size=size, dtype=torch.int64),
    )
    buffers = {key: [] for key in specs}
    for _ in range(flags.num_buffers):
        for key in buffers:
            buffers[key].append(torch.empty(**specs[key]).share_memory_())
    return buffers


def _format_observations(observation, keys=("glyphs", "status")):
    observations = {}
    for key in keys:
        entry = observation[key]
        entry = torch.from_numpy(entry)
        entry = entry.view((1, 1) + entry.shape)  # (...) -> (T,B,...).
        observations[key] = entry
    return observations


class ResettingEnvironment:
    """Turns a Gym environment into something that can be step()ed indefinitely."""

    def __init__(self, gym_env):
        self.gym_env = gym_env
        self.episode_return = None
        self.episode_step = None

    def initial(self):
        initial_reward = torch.zeros(1, 1)
        # This supports only single-tensor actions ATM.
        initial_last_action = torch.zeros(1, 1, dtype=torch.int64)
        self.episode_return = torch.zeros(1, 1)
        self.episode_step = torch.zeros(1, 1, dtype=torch.int32)
        initial_done = torch.ones(1, 1, dtype=torch.uint8)

        result = _format_observations(self.gym_env.reset())
        result.update(
            reward=initial_reward,
            done=initial_done,
            episode_return=self.episode_return,
            episode_step=self.episode_step,
            last_action=initial_last_action,
        )
        return result

    def step(self, action):
        observation, reward, done, unused_info = self.gym_env.step(action.item())
        self.episode_step += 1
        self.episode_return += reward
        episode_step = self.episode_step
        episode_return = self.episode_return
        if done:
            observation = self.gym_env.reset()
            self.episode_return = torch.zeros(1, 1)
            self.episode_step = torch.zeros(1, 1, dtype=torch.int32)

        result = _format_observations(observation)

        reward = torch.tensor(reward).view(1, 1)
        done = torch.tensor(done).view(1, 1)

        result.update(
            reward=reward,
            done=done,
            episode_return=episode_return,
            episode_step=episode_step,
            last_action=action,
        )
        return result

    def close(self):
        self.gym_env.close()


def train(flags):  # pylint: disable=too-many-branches, too-many-statements
    flags.savedir = os.path.expandvars(os.path.expanduser(flags.savedir))

    rundir = os.path.join(
        flags.savedir, "torchbeast-%s" % time.strftime("%Y%m%d-%H%M%S")
    )

    if not os.path.exists(rundir):
        os.makedirs(rundir)
    logging.info("Logging results to %s", rundir)

    symlink = os.path.join(flags.savedir, "latest")
    try:
        if os.path.islink(symlink):
            os.remove(symlink)
        if not os.path.exists(symlink):
            os.symlink(rundir, symlink)
        logging.info("Symlinked log directory: %s", symlink)
    except OSError:
        raise

    logfile = open(os.path.join(rundir, "logs.tsv"), "a", buffering=1)
    checkpointpath = os.path.join(rundir, "model.tar")

    flags.rundir = rundir

    if flags.num_buffers is None:  # Set sensible default for num_buffers.
        flags.num_buffers = max(2 * flags.num_actors, flags.batch_size)
    if flags.num_actors >= flags.num_buffers:
        raise ValueError("num_buffers should be larger than num_actors")
    if flags.num_buffers < flags.batch_size:
        raise ValueError("num_buffers should be larger than batch_size")

    T = flags.unroll_length
    B = flags.batch_size

    flags.device = None
    if not flags.disable_cuda and torch.cuda.is_available():
        logging.info("Using CUDA.")
        flags.device = torch.device("cuda")
    else:
        logging.info("Not using CUDA.")
        flags.device = torch.device("cpu")

    env = create_env(flags.env, archivefile=None)
    observation_space = env.observation_space
    action_space = env.action_space
    del env  # End this before forking.

    model = Net(observation_space, action_space.n, flags.use_lstm)
    buffers = create_buffers(flags, observation_space, model.num_actions)

    model.share_memory()

    # Add initial RNN state.
    initial_agent_state_buffers = []
    for _ in range(flags.num_buffers):
        state = model.initial_state(batch_size=1)
        for t in state:
            t.share_memory_()
        initial_agent_state_buffers.append(state)

    actor_processes = []
    ctx = mp.get_context("fork")
    free_queue = ctx.SimpleQueue()
    full_queue = ctx.SimpleQueue()

    for i in range(flags.num_actors):
        actor = ctx.Process(
            target=act,
            args=(
                flags,
                i,
                free_queue,
                full_queue,
                model,
                buffers,
                initial_agent_state_buffers,
            ),
            name="Actor-%i" % i,
        )
        actor.start()
        actor_processes.append(actor)

    learner_model = Net(observation_space, action_space.n, flags.use_lstm).to(
        device=flags.device
    )
    learner_model.load_state_dict(model.state_dict())

    optimizer = torch.optim.RMSprop(
        learner_model.parameters(),
        lr=flags.learning_rate,
        momentum=flags.momentum,
        eps=flags.epsilon,
        alpha=flags.alpha,
    )

    def lr_lambda(epoch):
        return 1 - min(epoch * T * B, flags.total_steps) / flags.total_steps

    scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda)

    stat_keys = [
        "total_loss",
        "mean_episode_return",
        "pg_loss",
        "baseline_loss",
        "entropy_loss",
    ]
    logfile.write("# Step\t%s\n" % "\t".join(stat_keys))

    step, stats = 0, {}

    def batch_and_learn(i, lock=threading.Lock()):
        """Thread target for the learning process."""
        nonlocal step, stats
        while step < flags.total_steps:
            batch, agent_state = get_batch(
                flags, free_queue, full_queue, buffers, initial_agent_state_buffers
            )
            stats = learn(
                flags, model, learner_model, batch, agent_state, optimizer, scheduler
            )
            with lock:
                logfile.write("%i\t" % step)
                logfile.write("\t".join(str(stats[k]) for k in stat_keys))
                logfile.write("\n")
                step += T * B

    for m in range(flags.num_buffers):
        free_queue.put(m)

    threads = []
    for i in range(flags.num_learner_threads):
        thread = threading.Thread(
            target=batch_and_learn,
            name="batch-and-learn-%d" % i,
            args=(i,),
            daemon=True,  # To support KeyboardInterrupt below.
        )
        thread.start()
        threads.append(thread)

    def checkpoint():
        if flags.disable_checkpoint:
            return
        logging.info("Saving checkpoint to %s", checkpointpath)
        torch.save(
            {
                "model_state_dict": model.state_dict(),
                "optimizer_state_dict": optimizer.state_dict(),
                "scheduler_state_dict": scheduler.state_dict(),
                "flags": vars(flags),
            },
            checkpointpath,
        )

    timer = timeit.default_timer
    try:
        last_checkpoint_time = timer()
        while step < flags.total_steps:
            start_step = step
            start_time = timer()
            time.sleep(5)

            if timer() - last_checkpoint_time > 10 * 60:  # Save every 10 min.
                checkpoint()
                last_checkpoint_time = timer()

            sps = (step - start_step) / (timer() - start_time)
            if stats.get("episode_returns", None):
                mean_return = (
                    "Return per episode: %.1f. " % stats["mean_episode_return"]
                )
            else:
                mean_return = ""
            total_loss = stats.get("total_loss", float("inf"))
            logging.info(
                "Steps %i @ %.1f SPS. Loss %f. %sStats:\n%s",
                step,
                sps,
                total_loss,
                mean_return,
                pprint.pformat(stats),
            )
    except KeyboardInterrupt:
        logging.warning("Quitting.")
        return  # Try joining actors then quit.
    else:
        for thread in threads:
            thread.join()
        logging.info("Learning finished after %d steps.", step)
    finally:
        for _ in range(flags.num_actors):
            free_queue.put(None)
        for actor in actor_processes:
            actor.join(timeout=1)

    checkpoint()
    logfile.close()


def test(flags, num_episodes=10):
    flags.savedir = os.path.expandvars(os.path.expanduser(flags.savedir))
    checkpointpath = os.path.join(flags.savedir, "latest", "model.tar")

    gym_env = create_env(flags.env, archivefile=None)
    env = ResettingEnvironment(gym_env)
    model = Net(gym_env.observation_space, gym_env.action_space.n, flags.use_lstm)
    model.eval()
    checkpoint = torch.load(checkpointpath, map_location="cpu")
    model.load_state_dict(checkpoint["model_state_dict"])

    observation = env.initial()
    returns = []

    agent_state = model.initial_state(batch_size=1)

    while len(returns) < num_episodes:
        if flags.mode == "test_render":
            env.gym_env.render()
        policy_outputs, agent_state = model(observation, agent_state)
        observation = env.step(policy_outputs["action"])
        if observation["done"].item():
            returns.append(observation["episode_return"].item())
            logging.info(
                "Episode ended after %d steps. Return: %.1f",
                observation["episode_step"].item(),
                observation["episode_return"].item(),
            )
    env.close()
    logging.info(
        "Average returns over %i steps: %.1f", num_episodes, sum(returns) / len(returns)
    )


class RandomNet(nn.Module):
    def __init__(self, observation_shape, num_actions, use_lstm):
        super(RandomNet, self).__init__()
        del observation_shape, use_lstm
        self.num_actions = num_actions
        self.theta = torch.nn.Parameter(torch.zeros(self.num_actions))

    def forward(self, inputs, core_state):
        # print(inputs)
        T, B, *_ = inputs["observation"].shape
        zeros = self.theta * 0
        # set logits to 0
        policy_logits = zeros[None, :].expand(T * B, -1)
        # set baseline to 0
        baseline = policy_logits.sum(dim=1).view(-1, B)

        # sample random action
        action = torch.multinomial(F.softmax(policy_logits, dim=1), num_samples=1).view(
            T, B
        )
        policy_logits = policy_logits.view(T, B, self.num_actions)
        return (
            dict(policy_logits=policy_logits, baseline=baseline, action=action),
            core_state,
        )

    def initial_state(self, batch_size):
        return ()


def _step_to_range(delta, num_steps):
    """Range of `num_steps` integers with distance `delta` centered around zero."""
    return delta * torch.arange(-num_steps // 2, num_steps // 2)


class Crop(nn.Module):
    """Helper class for NetHackNet below."""

    def __init__(self, height, width, height_target, width_target):
        super(Crop, self).__init__()
        self.width = width
        self.height = height
        self.width_target = width_target
        self.height_target = height_target
        width_grid = _step_to_range(2 / (self.width - 1), self.width_target)[
            None, :
        ].expand(self.height_target, -1)
        height_grid = _step_to_range(2 / (self.height - 1), height_target)[
            :, None
        ].expand(-1, self.width_target)

        # "clone" necessary, https://github.com/pytorch/pytorch/issues/34880
        self.register_buffer("width_grid", width_grid.clone())
        self.register_buffer("height_grid", height_grid.clone())

    def forward(self, inputs, coordinates):
        """Calculates centered crop around given x,y coordinates.
        Args:
           inputs [B x H x W]
           coordinates [B x 2] x,y coordinates
        Returns:
           [B x H' x W'] inputs cropped and centered around x,y coordinates.
        """
        assert inputs.shape[1] == self.height
        assert inputs.shape[2] == self.width

        inputs = inputs[:, None, :, :].float()

        x = coordinates[:, 0]
        y = coordinates[:, 1]

        x_shift = 2 / (self.width - 1) * (x.float() - self.width // 2)
        y_shift = 2 / (self.height - 1) * (y.float() - self.height // 2)

        grid = torch.stack(
            [
                self.width_grid[None, :, :] + x_shift[:, None, None],
                self.height_grid[None, :, :] + y_shift[:, None, None],
            ],
            dim=3,
        )

        # TODO: only cast to int if original tensor was int
        return (
            torch.round(F.grid_sample(inputs, grid, align_corners=True))
            .squeeze(1)
            .long()
        )


class NetHackNet(nn.Module):
    def __init__(
        self,
        observation_shape,
        num_actions,
        use_lstm,
        embedding_dim=32,
        crop_dim=9,
        num_layers=5,
    ):
        super(NetHackNet, self).__init__()

        self.glyph_shape = observation_shape["glyphs"].shape
        self.status_size = observation_shape["status"].shape[0]

        self.num_actions = num_actions
        self.use_lstm = use_lstm

        self.H = self.glyph_shape[0]
        self.W = self.glyph_shape[1]

        self.k_dim = embedding_dim
        self.h_dim = 512

        self.crop_dim = crop_dim

        self.crop = Crop(self.H, self.W, self.crop_dim, self.crop_dim)

        self.embed = nn.Embedding(5991, self.k_dim)

        K = embedding_dim  # number of input filters
        F = 3  # filter dimensions
        S = 1  # stride
        P = 1  # padding
        M = 16  # number of intermediate filters
        Y = 8  # number of output filters
        L = num_layers  # number of convnet layers

        in_channels = [K] + [M] * (L - 1)
        out_channels = [M] * (L - 1) + [Y]

        def interleave(xs, ys):
            return [val for pair in zip(xs, ys) for val in pair]

        conv_extract = [
            nn.Conv2d(
                in_channels=in_channels[i],
                out_channels=out_channels[i],
                kernel_size=(F, F),
                stride=S,
                padding=P,
            )
            for i in range(L)
        ]

        self.extract_representation = nn.Sequential(
            *interleave(conv_extract, [nn.ELU()] * len(conv_extract))
        )

        # CNN crop model.
        conv_extract_crop = [
            nn.Conv2d(
                in_channels=in_channels[i],
                out_channels=out_channels[i],
                kernel_size=(F, F),
                stride=S,
                padding=P,
            )
            for i in range(L)
        ]

        self.extract_crop_representation = nn.Sequential(
            *interleave(conv_extract_crop, [nn.ELU()] * len(conv_extract))
        )

        out_dim = self.k_dim
        # CNN over full glyph map
        out_dim += self.H * self.W * Y

        # CNN crop model.
        out_dim += self.crop_dim ** 2 * Y

        self.embed_status = nn.Sequential(
            nn.Linear(self.status_size, self.k_dim),
            nn.ReLU(),
            nn.Linear(self.k_dim, self.k_dim),
            nn.ReLU(),
        )

        self.fc = nn.Sequential(
            nn.Linear(out_dim, self.h_dim),
            nn.ReLU(),
            nn.Linear(self.h_dim, self.h_dim),
            nn.ReLU(),
        )

        if self.use_lstm:
            self.core = nn.LSTM(self.h_dim, self.h_dim, num_layers=1)

        self.policy = nn.Linear(self.h_dim, self.num_actions)
        self.baseline = nn.Linear(self.h_dim, 1)

    def initial_state(self, batch_size=1):
        if not self.use_lstm:
            return tuple()
        return tuple(
            torch.zeros(self.core.num_layers, batch_size, self.core.hidden_size)
            for _ in range(2)
        )

    def _select(self, embed, x):
        # Work around slow backward pass of nn.Embedding, see
        # https://github.com/pytorch/pytorch/issues/24912
        out = embed.weight.index_select(0, x.reshape(-1))
        return out.reshape(x.shape + (-1,))

    def forward(self, env_outputs, core_state):
        # -- [T x B x H x W]
        glyphs = env_outputs["glyphs"]

        # -- [T x B x F]
        status = env_outputs["status"]

        T, B, *_ = glyphs.shape

        # -- [B' x H x W]
        glyphs = torch.flatten(glyphs, 0, 1)  # Merge time and batch.

        # -- [B' x F]
        status = status.view(T * B, -1).float()

        # -- [B x H x W]
        glyphs = glyphs.long()
        # -- [B x 2] x,y coordinates
        coordinates = status[:, :2]
        # TODO ???
        # coordinates[:, 0].add_(-1)

        # -- [B x F]
        # FIXME: hack to use compatible status to before
        # status = status[:, [0, 1, 21, 10, 11]]

        status = status.view(T * B, -1).float()
        # -- [B x K]
        status_emb = self.embed_status(status)

        assert status_emb.shape[0] == T * B

        reps = [status_emb]

        # -- [B x H' x W']
        crop = self.crop(glyphs, coordinates)

        # print("crop", crop)
        # print("at_xy", glyphs[:, coordinates[:, 1].long(), coordinates[:, 0].long()])

        # -- [B x H' x W' x K]
        crop_emb = self._select(self.embed, crop)

        # CNN crop model.
        # -- [B x K x W' x H']
        crop_emb = crop_emb.transpose(1, 3)  # -- TODO: slow?
        # -- [B x W' x H' x K]
        crop_rep = self.extract_crop_representation(crop_emb)

        # -- [B x K']
        crop_rep = crop_rep.view(T * B, -1)
        assert crop_rep.shape[0] == T * B

        reps.append(crop_rep)

        # -- [B x H x W x K]
        glyphs_emb = self._select(self.embed, glyphs)
        # glyphs_emb = self.embed(glyphs)
        # -- [B x K x W x H]
        glyphs_emb = glyphs_emb.transpose(1, 3)  # -- TODO: slow?
        # -- [B x W x H x K]
        glyphs_rep = self.extract_representation(glyphs_emb)

        # -- [B x K']
        glyphs_rep = glyphs_rep.view(T * B, -1)

        assert glyphs_rep.shape[0] == T * B

        # -- [B x K'']
        reps.append(glyphs_rep)

        st = torch.cat(reps, dim=1)

        # -- [B x K]
        st = self.fc(st)

        if self.use_lstm:
            core_input = st.view(T, B, -1)
            core_output_list = []
            notdone = (~env_outputs["done"]).float()
            for input, nd in zip(core_input.unbind(), notdone.unbind()):
                # Reset core state to zero whenever an episode ended.
                # Make `done` broadcastable with (num_layers, B, hidden_size)
                # states:
                nd = nd.view(1, -1, 1)
                core_state = tuple(nd * s for s in core_state)
                output, core_state = self.core(input.unsqueeze(0), core_state)
                core_output_list.append(output)
            core_output = torch.flatten(torch.cat(core_output_list), 0, 1)
        else:
            core_output = st

        # -- [B x A]
        policy_logits = self.policy(core_output)
        # -- [B x A]
        baseline = self.baseline(core_output)

        if self.training:
            action = torch.multinomial(F.softmax(policy_logits, dim=1), num_samples=1)
        else:
            # Don't sample when testing.
            action = torch.argmax(policy_logits, dim=1)

        policy_logits = policy_logits.view(T, B, self.num_actions)
        baseline = baseline.view(T, B)
        action = action.view(T, B)

        return (
            dict(policy_logits=policy_logits, baseline=baseline, action=action),
            core_state,
        )


Net = NetHackNet


def main(flags):
    if flags.mode == "train":
        train(flags)
    else:
        test(flags)


if __name__ == "__main__":
    flags = parser.parse_args()
    main(flags)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Observation(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsObservation(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Observation()
        x.Init(buf, n + offset)
        return x

    # Observation
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Observation
    def Glyphs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from nle.fbs.NDArray import NDArray
            obj = NDArray()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Observation
    def Chars(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from nle.fbs.NDArray import NDArray
            obj = NDArray()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Observation
    def Colors(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from nle.fbs.NDArray import NDArray
            obj = NDArray()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Observation
    def Specials(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from nle.fbs.NDArray import NDArray
            obj = NDArray()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Observation
    def Status(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from nle.fbs.Status import Status
            obj = Status()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Observation
    def Inventory(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from nle.fbs.InventoryItem import InventoryItem
            obj = InventoryItem()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Observation
    def InventoryLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Observation
    def InventoryIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

def ObservationStart(builder): builder.StartObject(6)
def ObservationAddGlyphs(builder, glyphs): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(glyphs), 0)
def ObservationAddChars(builder, chars): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(chars), 0)
def ObservationAddColors(builder, colors): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(colors), 0)
def ObservationAddSpecials(builder, specials): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(specials), 0)
def ObservationAddStatus(builder, status): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(status), 0)
def ObservationAddInventory(builder, inventory): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(inventory), 0)
def ObservationStartInventoryVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ObservationEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class DLevel(object):
    __slots__ = ['_tab']

    # DLevel
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DLevel
    def Dnum(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # DLevel
    def Dlevel(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(1))

def CreateDLevel(builder, dnum, dlevel):
    builder.Prep(1, 2)
    builder.PrependInt8(dlevel)
    builder.PrependInt8(dnum)
    return builder.Offset()



# Copyright (c) Facebook, Inc. and its affiliates.
import enum


def M(c):
    if isinstance(c, str):
        c = ord(c)
    return 0x80 | c


def C(c):
    if isinstance(c, str):
        c = ord(c)
    return 0x1F & c


# Missing here:
#   Some characters for text input (e.g., +).
# General menu handling isn't part of this either.


class CompassCardinalDirection(enum.IntEnum):
    N = ord("k")
    E = ord("l")
    S = ord("j")
    W = ord("h")


class CompassIntercardinalDirection(enum.IntEnum):
    NE = ord("u")
    SE = ord("n")
    SW = ord("b")
    NW = ord("y")


# Merge the above.
CompassDirection = enum.IntEnum(
    "CompassDirection",
    {
        **CompassCardinalDirection.__members__,
        **CompassIntercardinalDirection.__members__,
    },
)


class CompassCardinalDirectionLonger(enum.IntEnum):
    N = ord("K")
    E = ord("L")
    S = ord("J")
    W = ord("H")


class CompassIntercardinalDirectionLonger(enum.IntEnum):
    NE = ord("U")
    SE = ord("N")
    SW = ord("B")
    NW = ord("Y")


CompassDirectionLonger = enum.IntEnum(
    "CompassDirectionLonger",
    {
        **CompassCardinalDirectionLonger.__members__,
        **CompassIntercardinalDirectionLonger.__members__,
    },
)


class MiscDirection(enum.IntEnum):
    UP = ord("<")  # go up a staircase
    DOWN = ord(">")  # go down a staircase
    WAIT = ord(".")  # rest one move while doing nothing / apply to self


class MiscAction(enum.IntEnum):
    MORE = ord("\r")


class Command(enum.IntEnum):
    EXTCMD = ord("#")  # perform an extended command
    EXTLIST = M("?")  # list all extended commands
    ADJUST = M("a")  # adjust inventory letters
    ANNOTATE = M("A")  # name current level
    APPLY = ord("a")  # apply (use) a tool (pick-axe, key, lamp...)
    ATTRIBUTES = C("x")  # show your attributes
    AUTOPICKUP = ord("@")  # toggle the pickup option on/off
    CALL = ord("C")  # call (name) something
    CAST = ord("Z")  # zap (cast) a spell
    CHAT = M("c")  # talk to someone
    CLOSE = ord("c")  # close a door
    CONDUCT = M("C")  # list voluntary challenges you have maintained
    DIP = M("d")  # dip an object into something
    DROP = ord("d")  # drop an item
    DROPTYPE = ord("D")  # drop specific item types
    EAT = ord("e")  # eat something
    ENGRAVE = ord("E")  # engrave writing on the floor
    ENHANCE = M("e")  # advance or check weapon and spell skills
    FIRE = ord("f")  # fire ammunition from quiver
    FORCE = M("f")  # force a lock
    GLANCE = ord(";")  # show what type of thing a map symbol corresponds to
    HELP = ord("?")  # give a help message
    HISTORY = ord("V")  # show long version and game history
    INVENTORY = ord("i")  # show your inventory
    INVENTTYPE = ord("I")  # inventory specific item types
    INVOKE = M("i")  # invoke an object's special powers
    JUMP = M("j")  # jump to another location
    KICK = C("d")  # kick something
    KNOWN = ord("\\")  # show what object types have been discovered
    KNOWNCLASS = ord("`")  # show discovered types for one class of objects
    LOOK = ord(":")  # look at what is here
    LOOT = M("l")  # loot a box on the floor
    MONSTER = M("m")  # use monster's special ability
    OFFER = M("o")  # offer a sacrifice to the gods
    OPEN = ord("o")  # open a door
    OPTIONS = ord("O")  # show option settings, possibly change them
    OVERVIEW = C("o")  # show a summary of the explored dungeon
    PAY = ord("p")  # pay your shopping bill
    PICKUP = ord(",")  # pick up things at the current location
    PRAY = M("p")  # pray to the gods for help
    PREVMSG = C("p")  # view recent game messages
    PUTON = ord("P")  # put on an accessory (ring, amulet, etc)
    QUAFF = ord("q")  # quaff (drink) something
    QUIT = M("q")  # exit without saving current game
    QUIVER = ord("Q")  # select ammunition for quiver
    READ = ord("r")  # read a scroll or spellbook
    REDRAW = C("r")  # redraw screen
    REMOVE = ord("R")  # remove an accessory (ring, amulet, etc)
    RIDE = M("R")  # mount or dismount a saddled steed
    RUB = M("r")  # rub a lamp or a stone
    SAVE = ord("S")  # save the game and exit
    SEARCH = ord("s")  # search for traps and secret doors
    SEEALL = ord("*")  # show all equipment in use
    SEETRAP = ord("^")  # show the type of adjacent trap
    SIT = M("s")  # sit down
    SWAP = ord("x")  # swap wielded and secondary weapons
    TAKEOFF = ord("T")  # take off one piece of armor
    TAKEOFFALL = ord("A")  # remove all armor
    TELEPORT = C("t")  # teleport around the level
    THROW = ord("t")  # throw something
    TIP = M("T")  # empty a container
    TRAVEL = ord("_")  # travel to a specific location on the map
    TURN = M("t")  # turn undead away
    TWOWEAPON = ord("X")  # toggle two-weapon combat
    UNTRAP = M("u")  # untrap something
    VERSION = M("v")  # list compile time options for this version of NetHack
    VERSIONSHORT = ord("v")  # show version
    WEAR = ord("W")  # wear a piece of armor
    WHATDOES = ord("&")  # tell what a command does
    WHATIS = ord("/")  # show what type of thing a symbol corresponds to
    WIELD = ord("w")  # wield (put in use) a weapon
    WIPE = M("w")  # wipe off your face
    ZAP = ord("z")  # zap a wand


ACTIONS = tuple(
    list(CompassDirection)
    + list(CompassDirectionLonger)
    + list(MiscDirection)
    + list(MiscAction)
    + list(Command)
)

_ACTIONS_DICT = {}
for enum_class in [
    CompassDirection,
    CompassDirectionLonger,
    MiscDirection,
    MiscAction,
    Command,
]:
    for action in enum_class:
        _ACTIONS_DICT[action.value] = "%s.%s" % (enum_class.__name__, action.name)


def action_id_to_type(action):
    return _ACTIONS_DICT[action]



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ProgramState(object):
    __slots__ = ['_tab']

    # ProgramState
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ProgramState
    def Gameover(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # ProgramState
    def Panicking(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(1))
    # ProgramState
    def Exiting(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(2))
    # ProgramState
    def InMoveloop(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(3))
    # ProgramState
    def InImpossible(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4))

def CreateProgramState(builder, gameover, panicking, exiting, inMoveloop, inImpossible):
    builder.Prep(1, 5)
    builder.PrependBool(inImpossible)
    builder.PrependBool(inMoveloop)
    builder.PrependBool(exiting)
    builder.PrependBool(panicking)
    builder.PrependBool(gameover)
    return builder.Offset()
